from garaga.definitions import (
    CURVES,
    CurveID,
    G1G2Pair,
    G1Point,
    G2Point,
    get_base_field,
    get_sparsity,
)
from garaga.hints.frobenius import get_frobenius_maps
from garaga.hints.multi_miller_witness import get_final_exp_witness
from garaga.hints.tower_backup import E6, E12
from garaga.modulo_circuit import ModuloCircuitElement, PyFelt, WriteOps
from garaga.precompiled_circuits.multi_miller_loop import MultiMillerLoopCircuit


def get_root_and_scaling_factor(
    curve_id: int,
    P: list[G1Point | tuple[ModuloCircuitElement, ModuloCircuitElement]],
    Q: list[
        G2Point
        | tuple[
            tuple[ModuloCircuitElement, ModuloCircuitElement],
            tuple[ModuloCircuitElement, ModuloCircuitElement],
        ]
    ],
    m: list[ModuloCircuitElement] = None,
) -> tuple[list[PyFelt], list[PyFelt], list[int]]:
    assert (
        len(P) == len(Q) >= 2
    ), f"P and Q must have the same length and >= 2, got {len(P)} and {len(Q)}"
    field = get_base_field(curve_id)
    c_input: list[PyFelt] = []

    c: MultiMillerLoopCircuit = MultiMillerLoopCircuit(
        name="mock", curve_id=curve_id, n_pairs=len(P)
    )
    if isinstance(P[0], G1Point):
        c.write_p_and_q(P, Q)
    elif isinstance(P[0], tuple) and isinstance(P[0][0], ModuloCircuitElement):
        for p, q in zip(P, Q):
            c_input.append(p[0].felt)
            c_input.append(p[1].felt)
            c_input.append(q[0][0].felt)
            c_input.append(q[0][1].felt)
            c_input.append(q[1][0].felt)
            c_input.append(q[1][1].felt)
        c.write_p_and_q_raw(c_input)

    f = E12.from_direct(c.miller_loop(len(P)), curve_id)
    if m is not None:
        M = E12.from_direct(m, curve_id)
        f = f * M
    # h = (CURVES[curve_id].p ** 12 - 1) // CURVES[curve_id].n
    # assert f**h == E12.one(curve_id)
    lambda_root_e12, scaling_factor_e12 = get_final_exp_witness(curve_id, f)

    lambda_root: list[PyFelt]
    scaling_factor: list[PyFelt]

    lambda_root, scaling_factor = (
        (
            lambda_root_e12.__inv__().to_direct()
            if curve_id == CurveID.BLS12_381.value
            else lambda_root_e12.to_direct()
        ),  # Pass lambda_root inverse directly for BLS.
        scaling_factor_e12.to_direct(),
    )

    e6_subfield = E12([E6.random(curve_id), E6.zero(curve_id)], curve_id)
    scaling_factor_sparsity = get_sparsity(e6_subfield.to_direct())

    # Assert sparsity is correct: for every index where the sparsity is 0, the coefficient must 0 in scaling factor
    for i in range(len(scaling_factor_sparsity)):
        if scaling_factor_sparsity[i] == 0:
            assert scaling_factor[i].value == 0
    # Therefore scaling factor lies in Fp6

    return lambda_root, scaling_factor, scaling_factor_sparsity


def get_max_Q_degree(curve_id: int, n_pairs: int) -> int:
    if curve_id == CurveID.BN254.value:
        line_degree = 9
    elif curve_id == CurveID.BLS12_381.value:
        line_degree = 8
    else:
        raise NotImplementedError(f"Curve {curve_id} not implemented")

    f_degree = 11
    lambda_root_degree = 11

    if curve_id == CurveID.BN254.value:
        # Largest degree happen in bit_10 case where we do (f*f*C * Π_n_pairs(line)^2 * Π_n_pairs(line))
        max_q_degree = (
            4 * f_degree
            + 2 * lambda_root_degree
            + 4 * line_degree * n_pairs
            + line_degree * n_pairs
            - 12
        )
    else:
        # Largest Q happens in bit_00 case where we do (f*f* Π_n_pairs(line)^2 * Π_n_pairs(line)
        max_q_degree = (
            4 * f_degree + 2 * line_degree * n_pairs + line_degree * n_pairs - 12
        )

    return max_q_degree


class MultiPairingCheckCircuit(MultiMillerLoopCircuit):
    def __init__(
        self,
        name: str,
        curve_id: int,
        n_pairs: int,
        hash_input: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
        precompute_lines: bool = False,
        n_points_precomputed_lines: int = None,
    ):
        assert n_pairs >= 2, "n_pairs must be >= 2 for pairing checks"
        super().__init__(
            name=name,
            curve_id=curve_id,
            n_pairs=n_pairs,
            hash_input=hash_input,
            init_hash=init_hash,
            compilation_mode=compilation_mode,
            precompute_lines=precompute_lines,
            n_points_precomputed_lines=n_points_precomputed_lines,
        )
        self.frobenius_maps = {}
        for i in [1, 2, 3]:
            _, self.frobenius_maps[i] = get_frobenius_maps(
                curve_id=curve_id, extension_degree=self.extension_degree, frob_power=i
            )

    def frobenius(
        self, X: list[ModuloCircuitElement], frob_power: int
    ) -> list[ModuloCircuitElement]:
        frob = [None] * self.extension_degree
        for i, list_op in enumerate(self.frobenius_maps[frob_power]):
            list_op_result = []
            for index, constant in list_op:
                if constant == 1:
                    list_op_result.append(X[index])
                else:
                    list_op_result.append(
                        self.mul(
                            X[index], self.set_or_get_constant(self.field(constant))
                        )
                    )
            frob[i] = list_op_result[0]
            for op_res in list_op_result[1:]:
                frob[i] = self.add(frob[i], op_res)

        return frob

    def bit_0_case(
        self,
        f: list[ModuloCircuitElement],
        points: list[tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]],
        n_pairs: int,
    ):
        """
        Compute the bit 0 case of the Miller loop.
        params : f : the current miller loop FP12 element
                points : the list of points to double
                n_pairs : the number of pairs to double
        returns : the new miller loop FP12 element and the new points
        """
        assert len(points) == n_pairs
        new_lines = []
        new_points = []
        for k in range(n_pairs):
            T, l1 = self.double_step(points[k], k)
            new_lines.append(l1)
            new_points.append(T)

        # Square f and multiply by lines for all pairs
        new_f = self.extf_mul(
            [f, f, *new_lines],
            12,
            Ps_sparsities=[None, None] + [self.line_sparsity] * n_pairs,
        )
        return new_f, new_points

    def bit_00_case(
        self,
        f: list[ModuloCircuitElement],
        points: list[tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]],
        n_pairs: int,
    ):
        """
        Compute the bit 00 case of the Miller loop.
        params : f : the current miller loop FP12 element
                points : the list of points to double
                n_pairs : the number of pairs to double
        returns : the new miller loop FP12 element and the new points
        """
        assert len(points) == n_pairs
        new_lines = []
        new_points = []
        for k in range(n_pairs):
            T, l1 = self.double_step(points[k], k)
            new_lines.append(l1)
            new_lines.append(l1)  # Double since it's going to be squared
            new_points.append(T)

        new_new_points = []
        new_new_lines = []

        for k in range(n_pairs):
            T, l1 = self.double_step(new_points[k], k)
            new_new_lines.append(l1)
            new_new_points.append(T)

        # (f^2 * Π_(new_lines))^2 * Π_new_new_lines = f^4 * Π_new_lines^2 * Π_new_new_lines

        new_f = self.extf_mul(
            [f, f, f, f, *new_lines, *new_new_lines],
            12,
            Ps_sparsities=[None, None, None, None]
            + [self.line_sparsity] * n_pairs * 2
            + [self.line_sparsity] * n_pairs,
        )
        return new_f, new_new_points

    def bit_01_case(
        self,
        f: list[ModuloCircuitElement],
        points: list[tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]],
        Q_selects: list[tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]],
        c_or_c_inv: list[ModuloCircuitElement],
        n_pairs: int,
    ):

        assert len(points) == n_pairs
        new_lines = []
        new_points = []
        for k in range(n_pairs):
            T, l1 = self.double_step(points[k], k)
            new_lines.append(l1)
            new_lines.append(l1)  # Double since it's going to be squared
            new_points.append(T)

        new_new_points = []
        new_new_lines = []

        for k in range(n_pairs):
            T, l1, l2 = self.double_and_add_step(new_points[k], Q_selects[k], k)
            new_new_lines.append(l1)
            new_new_lines.append(l2)
            new_new_points.append(T)

        # (f^2 * Π_(new_lines))^2 * Π_new_new_lines = f^4 * Π_new_lines^2 * Π_new_new_lines
        new_f = self.extf_mul(
            [f, f, f, f, *new_lines, *new_new_lines, c_or_c_inv],
            12,
            Ps_sparsities=[None] * 4
            + [self.line_sparsity] * n_pairs * 2
            + [self.line_sparsity] * n_pairs * 2
            + [None],
        )
        return new_f, new_new_points

    def bit_10_case(
        self,
        f: list[ModuloCircuitElement],
        points: list[tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]],
        Q_selects: list[tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]],
        c_or_c_inv: list[ModuloCircuitElement],
        n_pairs: int,
    ):
        assert len(points) == n_pairs
        new_lines = []
        new_points = []
        for k in range(n_pairs):
            T, l1, l2 = self.double_and_add_step(points[k], Q_selects[k], k)
            new_lines.append(l1)
            new_lines.append(l1)  # Double since it's going to be squared
            new_lines.append(l2)
            new_lines.append(l2)  # Double since it's going to be squared
            new_points.append(T)

        new_new_points = []
        new_new_lines = []

        for k in range(n_pairs):
            T, l1 = self.double_step(new_points[k], k)
            new_new_lines.append(l1)
            new_new_points.append(T)

        new_f = self.extf_mul(
            [f, f, f, f, c_or_c_inv, c_or_c_inv, *new_lines, *new_new_lines],
            12,
            Ps_sparsities=[None] * 4
            + [None, None]
            + [self.line_sparsity] * n_pairs * 4
            + [self.line_sparsity] * n_pairs,
        )
        return new_f, new_new_points

    def bit_1_init_case(
        self,
        f: list[ModuloCircuitElement],
        points: list[tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]],
        n_pairs: int,
        c: list[ModuloCircuitElement],
    ):
        """
        Compute the bit 1 case of the Miller loop when it is the first bit (positive).
        Uses triple step instead of double and add.
        """
        assert len(points) == n_pairs
        new_lines = []
        new_points = []
        for k in range(n_pairs):
            T, l1, l2 = self.triple_step(points[k], k)
            new_lines.append(l1)
            new_lines.append(l2)
            new_points.append(T)

        # Square f and multiply by lines for all pairs
        new_f = self.extf_mul(
            [f, f, c, *new_lines],
            12,
            Ps_sparsities=[None, None, None] + [self.line_sparsity] * n_pairs * 2,
        )
        return new_f, new_points

    def bit_1_case(
        self,
        f: list[ModuloCircuitElement],
        points: list[tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]],
        Q_select: list[tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]],
        n_pairs: int,
        c_or_c_inv: list[ModuloCircuitElement],
    ):
        """
        Compute the bit 1 case of the Miller loop.
        params : f : the current miller loop FP12 element
                points : the list of points to double
                Q_select : the list of points to add.
                Q_select[k] is the point to add if the k-th bit is 1, and the negation of the point if the k-th bit is -1.
                n_pairs : the number of pairs to double
                c_or_c_inv : the lambda-th root c or its inverse depending on the bit (c_inv if 1, c if -1)
        returns : the new miller loop FP12 element and the new points
        """
        assert len(points) == n_pairs == len(Q_select)
        new_lines = []
        new_points = []
        for k in range(n_pairs):
            T, l1, l2 = self.double_and_add_step(points[k], Q_select[k], k)
            new_lines.append(l1)
            new_lines.append(l2)
            new_points.append(T)

        # Square f and multiply by lines for all pairs
        new_f = self.extf_mul(
            [f, f, c_or_c_inv, *new_lines],
            12,
            Ps_sparsities=[None, None, None] + [self.line_sparsity] * n_pairs * 2,
        )
        return new_f, new_points

    def multi_pairing_check(
        self, n_pairs: int, m: list[ModuloCircuitElement] | None = None
    ) -> tuple[
        list[ModuloCircuitElement],
        list[ModuloCircuitElement],
        list[ModuloCircuitElement],
    ]:

        lambda_root, scaling_factor, scaling_factor_sparsity = (
            get_root_and_scaling_factor(self.curve_id, self.P, self.Q, m)
        )

        c_or_c_inv = self.write_elements(lambda_root, WriteOps.COMMIT)

        w = self.write_elements(
            scaling_factor, WriteOps.COMMIT, scaling_factor_sparsity
        )

        if self.curve_id == CurveID.BLS12_381.value:
            # Conjugate c so that the final conjugate in BLS loop gives indeed f/c^(-x), as conjugate(f/conjugate(c^(-x))) = conjugate(f)/c^(-x)
            lambda_root = None
            lambda_root_inverse = c_or_c_inv
            c_inv = self.conjugate_e12d(lambda_root_inverse)
        elif self.curve_id == CurveID.BN254.value:
            lambda_root = c = c_or_c_inv
            lambda_root_inverse = c_inv = self.extf_inv(c_or_c_inv, 12)

        # Init f as 1/c = 1 / (λ-th √(f_output*scaling_factor)), where:
        # λ = 6 * x + 2 + q - q**2 + q**3 for BN
        # λ = -x + q for BLS
        # Miller loop will compute f_output/c^(6*x2) if BN, or f_output/c^(-x) if BLS

        f = c_inv

        start_index = len(self.loop_counter) - 2

        if self.loop_counter[start_index] == 1:
            # Handle case when first bit is +1, need to triple point instead of double and add.
            f, Qs = self.bit_1_init_case(f, self.Q, n_pairs, c_inv)
        elif self.loop_counter[start_index] == 0:
            f, Qs = self.bit_0_case(f, self.Q, n_pairs)
        else:
            raise NotImplementedError(
                f"Init bit {self.loop_counter[start_index]} not implemented"
            )

        i = start_index - 1
        while i >= 0:
            if self.loop_counter[i] == 0:  # First bit is 0
                if i > 0:  # Check next bit if it exists
                    next_bit = self.loop_counter[i - 1]
                    if next_bit == 0:
                        # 00 case
                        f, Qs = self.bit_00_case(f, Qs, n_pairs)
                        i -= 1  # Skip next bit
                    elif next_bit in (1, -1) and self.curve_id == CurveID.BN254.value:
                        # 01 or 0(-1) case
                        Q_selects = [
                            self.Q[k] if next_bit == 1 else self.Qneg[k]
                            for k in range(n_pairs)
                        ]
                        c_or_c_inv = c_inv if next_bit == 1 else c
                        f, Qs = self.bit_01_case(f, Qs, Q_selects, c_or_c_inv, n_pairs)
                        i -= 1  # Skip next bit
                    else:
                        # Single 0 (BLS only)
                        f, Qs = self.bit_0_case(f, Qs, n_pairs)
                else:
                    # Single 0 at the end
                    f, Qs = self.bit_0_case(f, Qs, n_pairs)

            elif self.loop_counter[i] in (1, -1):  # First bit is ±1
                # Calculate Q_selects and c_or_c_inv based on the si
                Q_selects = [
                    self.Q[k] if self.loop_counter[i] == 1 else self.Qneg[k]
                    for k in range(n_pairs)
                ]
                c_or_c_inv = c_inv if self.loop_counter[i] == 1 else c

                if (
                    i > 0
                    and self.loop_counter[i - 1] == 0
                    and self.curve_id == CurveID.BN254.value
                ):
                    # 10 or (-1)0 case
                    f, Qs = self.bit_10_case(f, Qs, Q_selects, c_or_c_inv, n_pairs)
                    i -= 1  # Skip next bit
                elif i == 0 or self.curve_id == CurveID.BLS12_381.value:
                    # Single ±1 at the end
                    f, Qs = self.bit_1_case(f, Qs, Q_selects, n_pairs, c_or_c_inv)
                else:
                    raise NotImplementedError(
                        f"Bit {self.loop_counter[i]} not implemented"
                    )
            else:
                raise NotImplementedError(f"Bit {self.loop_counter[i]} not implemented")
            i -= 1

        if m is not None and len(m) == 12:
            final_r_sparsity = None
        else:
            final_r_sparsity = [1] + [0] * 11

        if self.curve_id == CurveID.BN254.value:
            lines = self.bn254_finalize_step(Qs)
            f = self.extf_mul(
                [f, *lines],
                12,
                Ps_sparsities=[None] + [self.line_sparsity] * self.n_pairs * 2,
            )
            # λ = 6 * x + 2 + q - q**2 + q**3
            c_inv_frob_1 = self.frobenius(c_inv, 1)
            c_frob_2 = self.frobenius(c, 2)
            c_inv_frob_3 = self.frobenius(c_inv, 3)

            f = self.extf_mul(
                ([f, w, c_inv_frob_1, c_frob_2, c_inv_frob_3]),
                12,
                Ps_sparsities=([None, scaling_factor_sparsity, None, None, None]),
                r_sparsity=final_r_sparsity,
            )

        elif self.curve_id == CurveID.BLS12_381.value:
            # λ = -x + q for BLS
            c_inv_frob_1 = self.frobenius(c_inv, 1)
            f = self.extf_mul(
                Ps=[f, w, c_inv_frob_1],
                extension_degree=12,
                Ps_sparsities=[None, scaling_factor_sparsity, None],
                r_sparsity=final_r_sparsity,
            )
            if m is not None and len(m) == 12:
                f = self.conjugate_e12d(f)

        else:
            raise NotImplementedError(f"Curve {self.curve_id} not implemented")

        if m is not None and len(m) == 12:
            f = self.extf_mul([f, m], 12, r_sparsity=[1] + [0] * 11)

        assert [fi.value for fi in f] == [1] + [
            0
        ] * 11, f"Pairing check failed: {[fi.value for fi in f]}"
        return (
            f,
            lambda_root,
            lambda_root_inverse,
            scaling_factor,
            scaling_factor_sparsity,
        )


def get_pairing_check_input(
    curve_id: CurveID, n_pairs: int, include_m: bool = False, return_pairs: bool = False
) -> tuple[list[PyFelt | G1G2Pair], list[PyFelt] | G1G2Pair | None]:
    """
    Returns a list of G1G2Pairs and the extra public pair if include_m is True
    """
    n_pairs = n_pairs + include_m

    assert n_pairs >= 2, "n_pairs must be >= 2 for pairing checks"
    field = get_base_field(curve_id.value)
    if n_pairs == 2:
        # Generate inputs resembling BLS signature verification
        curve = CURVES[curve_id.value]
        secret_key = field.random(curve.n).value
        public_key = G2Point.get_nG(curve_id, secret_key)
        message_hash = G1Point.gen_random_point(curve_id)
        signature = message_hash.scalar_mul(secret_key)

        P = [signature, message_hash]
        Q = [G2Point.get_nG(curve_id, 1), -public_key]
    else:
        p = G1Point.gen_random_point(curve_id)
        q = G2Point.gen_random_point(curve_id)

        P = [p] * n_pairs
        Q = [q] * n_pairs

        P[-1] = p.scalar_mul(-(n_pairs - 1))
    c_input = []
    for p, q in zip(P, Q):
        c_input.append(field(p.x))
        c_input.append(field(p.y))
        c_input.append(field(q.x[0]))
        c_input.append(field(q.x[1]))
        c_input.append(field(q.y[0]))
        c_input.append(field(q.y[1]))

    if return_pairs:
        if include_m:
            return [G1G2Pair(p, q) for p, q in zip(P[:-1], Q[:-1])], G1G2Pair(
                P[-1], Q[-1]
            )
        else:
            return [G1G2Pair(p, q) for p, q in zip(P, Q)], None
    else:
        if include_m:
            mloop_circuit = MultiMillerLoopCircuit(
                name="mock", curve_id=curve_id.value, n_pairs=1
            )
            mloop_circuit.write_p_and_q_raw(c_input[-6:])
            M = mloop_circuit.miller_loop(n_pairs=1)
            M = [mi.felt for mi in M]
            return c_input[:-6], M
        else:
            return c_input, None



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import List, Union

from garaga.algebra import BaseField, ModuloCircuitElement, PyFelt
from garaga.definitions import BASE, CURVES, N_LIMBS, STARK, CurveID, get_sparsity
from garaga.hints.io import bigint_split
from garaga.modulo_circuit_structs import (
    Cairo1SerializableStruct,
    GenericT,
    u128,
    u256,
    u384,
)

BATCH_SIZE = 1  # Batch Size, only used in cairo 0 mode.


class WriteOps(Enum):
    """
    Enum for the source of a write operation in the value segment.
    -CONSTANT | INPUT: The value was written by coping N_LIMBS from the allocation pointer to the value segment.
    -COMMIT: The value was written as a result of a hint value from the prover, and needs to be hashed.
    -WITNESS: Same as COMMIT, but do not need to be hashed.
    -FELT: The value was written by coping N_LIMBS from the allocation pointer,
           as a result of a non-deterministic felt252 to UInt384 decomposition.
           In CairoZero, the value segment is increased by 1 due to an extra range check needed.
    -BUILTIN: The value was written by the modulo builtin as result of a ADD or MUL instruction.
    """

    CONSTANT = auto()
    INPUT = auto()
    COMMIT = auto()
    WITNESS = auto()
    FELT = auto()
    BUILTIN = auto()


class ModBuiltinOps(Enum):
    """
    Enum for the modulo builtin operations.
    """

    ADD = "+"
    MUL = "*"


class CairoVMOps(Enum):
    ASSERT_EQ = "=="


@dataclass(slots=True, frozen=True)
class ModuloCircuitInstruction:
    operation: ModBuiltinOps
    left_offset: int
    right_offset: int
    result_offset: int
    comment: str | None


@dataclass(slots=True, frozen=True)
class ValueSegmentItem:
    emulated_felt: PyFelt
    write_source: WriteOps
    instruction: None | ModuloCircuitInstruction

    @property
    def value(self) -> int:
        return self.emulated_felt.value

    @property
    def p(self):
        return self.emulated_felt.p

    @property
    def felt(self):
        return self.emulated_felt


@dataclass(slots=True, init=False)
class ValueSegment:
    segment: dict[int, ValueSegmentItem]
    assert_eq_instructions: list[ModuloCircuitInstruction]
    offset: int
    n_limbs: int
    debug: bool
    name: str
    output: list[ModuloCircuitElement]
    segment_stacks: dict[WriteOps, dict[int, ValueSegmentItem]]
    output_structs: list[Cairo1SerializableStruct] | None

    def __init__(self, name: str, debug: bool = False, compilation_mode: int = 0):
        self.segment: dict[int, ValueSegmentItem] = dict()
        self.assert_eq_instructions: list[ModuloCircuitInstruction] = []
        self.output: list[ModuloCircuitElement] = []
        self.offset = 0
        self.n_limbs = N_LIMBS
        self.debug = debug
        self.name = name
        self.segment_stacks = {key: dict() for key in WriteOps}
        self.output_structs = None if compilation_mode == 0 else []

    def __len__(self) -> int:
        return len(self.segment)

    def __getitem__(self, key: int) -> ValueSegmentItem:
        return self.segment[key]

    @property
    def input(self) -> list[ModuloCircuitElement]:
        combined_items = [
            (offset, item)
            for stack in [
                WriteOps.INPUT,
                WriteOps.COMMIT,
                WriteOps.WITNESS,
                WriteOps.FELT,
            ]
            for offset, item in self.segment_stacks[stack].items()
        ]
        combined_items.sort(key=lambda x: x[0])  # Sort by offset

        return [
            ModuloCircuitElement(item.emulated_felt, offset)
            for offset, item in combined_items
        ]

    def write_to_segment(self, item: ValueSegmentItem) -> int:
        offset = self.offset
        self.segment[offset] = item
        self.segment_stacks[item.write_source][offset] = item
        self.offset += N_LIMBS
        if item.write_source == WriteOps.FELT:
            self.offset += 1
        return offset

    def non_interactive_transform(self) -> "ValueSegment":
        """
        Rebuild a new ValueSegment by re-ordering the current one in this order:
        CONSTANT
        INPUT
        COMMIT
        WITNESS
        FELT
        BUILTIN
        Order matters!
        """
        res = ValueSegment(self.name)
        offset_map = {}
        for stacks_key in [
            WriteOps.CONSTANT,
            WriteOps.INPUT,
            WriteOps.COMMIT,
            WriteOps.WITNESS,
            WriteOps.FELT,
            WriteOps.BUILTIN,
        ]:
            if self.debug:
                print(stacks_key, len(self.segment_stacks[stacks_key]))
            for old_offset, item in self.segment_stacks[stacks_key].items():
                # print(offset_map)
                if item.instruction is None:
                    new_offset = res.write_to_segment(item)
                else:
                    # print(item.instruction)
                    new_offset = res.write_to_segment(
                        ValueSegmentItem(
                            item.emulated_felt,
                            item.write_source,
                            ModuloCircuitInstruction(
                                operation=item.instruction.operation,
                                left_offset=offset_map.get(
                                    item.instruction.left_offset, res.offset
                                ),
                                right_offset=offset_map.get(
                                    item.instruction.right_offset, res.offset
                                ),
                                result_offset=offset_map.get(
                                    item.instruction.result_offset, res.offset
                                ),
                                comment=item.instruction.comment,
                            ),
                        )
                    )

                offset_map[old_offset] = new_offset
        # Those are builtins instructions that did not create any new value.
        for assert_eq_instruction in self.assert_eq_instructions:
            res.assert_eq_instructions.append(
                ModuloCircuitInstruction(
                    operation=assert_eq_instruction.operation,
                    left_offset=offset_map[assert_eq_instruction.left_offset],
                    right_offset=offset_map[assert_eq_instruction.right_offset],
                    result_offset=offset_map[assert_eq_instruction.result_offset],
                    comment=assert_eq_instruction.comment,
                )
            )
        new_output = []
        new_output_structs = None if self.output_structs is None else []
        for elmt in self.output:
            new_output.append(ModuloCircuitElement(elmt.felt, offset_map[elmt.offset]))

        if self.output_structs is not None:
            for struct in self.output_structs:
                new_output_structs.append(
                    struct.__class__(
                        struct.name,
                        [
                            ModuloCircuitElement(elt.felt, offset_map[elt.offset])
                            for elt in struct.elmts
                        ],
                    )
                )

        res.output = new_output
        res.output_structs = new_output_structs
        return res

    def get_dw_lookups(self) -> dict:
        """
        Returns the DW arrays for the compiled circuit.
        Only relevant in Cairo 0 mode.
        """
        dw_arrays = {
            "constants_ptr": [],
            "add_offsets_ptr": [],
            "mul_offsets_ptr": [],
            "output_offsets_ptr": [],
            "poseidon_indexes_ptr": [],
        }
        for _, item in self.segment_stacks[WriteOps.CONSTANT].items():
            dw_arrays["constants_ptr"].append(
                bigint_split(item.value, self.n_limbs, BASE)
            )
        for write_op in [WriteOps.BUILTIN, WriteOps.WITNESS]:
            for result_offset, item in self.segment_stacks[write_op].items():
                if not item.instruction:
                    continue
                # print(item.instruction)
                dw_arrays[
                    item.instruction.operation.name.lower() + "_offsets_ptr"
                ].append(
                    (
                        item.instruction.left_offset,
                        item.instruction.right_offset,
                        item.instruction.result_offset,
                        item.instruction.comment,
                    )
                )
        for assert_eq_instruction in self.assert_eq_instructions:
            dw_arrays[
                assert_eq_instruction.operation.name.lower() + "_offsets_ptr"
            ].append(
                (
                    assert_eq_instruction.left_offset,
                    assert_eq_instruction.right_offset,
                    assert_eq_instruction.result_offset,
                    assert_eq_instruction.comment,
                )
            )
        for output_elmt in self.output:
            dw_arrays["output_offsets_ptr"].append(output_elmt.offset)

        # print("dw_arrays[add_offsets_ptr]", dw_arrays["add_offsets_ptr"])
        return dw_arrays

    def print(self):
        # ANSI escape codes for some colors
        RED = "\033[31m"  # Red text
        GREEN = "\033[32m"  # Green text
        BLUE = "\033[34m"  # Blue text
        ORANGE = "\033[33m"  # Orange text
        RESET = "\033[0m"  # Reset to default color

        # print(f"{RED}This will be red!{RESET}")
        # print(f"{GREEN}This will be green!{RESET}")

        for i, (offset, val) in enumerate(self.segment.items()):
            row = ""
            row += f"{BLUE}{'['+str(i)+']':^7}{RESET}|{ORANGE}{'['+str(offset)+']':^7}{RESET}|"
            if val.instruction:
                row += f"{str(val.instruction.left_offset)+str(val.instruction.operation.value)+str(val.instruction.right_offset)+'='+str(val.instruction.result_offset):^9}|"
            else:
                row += f"{'_':^9}|"
            row += f"{val.write_source.name:^8}|"
            import numpy as np

            row += f"\t{RED}{np.base_repr(val.emulated_felt.value, base=36)}{RESET} {val.instruction.comment if val.instruction else ''}"

            print(row)

    def summarize(self):
        all_instructions = {
            item.instruction for item in self.segment.values() if item.instruction
        }
        all_assert_eq_instructions = set(self.assert_eq_instructions)

        all_instructions_except_assert_eq = (
            all_instructions - all_assert_eq_instructions
        )
        add_count: int = 0
        mul_count: int = 0
        for instruction in all_instructions_except_assert_eq:
            if instruction.operation.name == "ADD":
                add_count += 1
            elif instruction.operation.name == "MUL":
                mul_count += 1
        return add_count, mul_count, len(all_assert_eq_instructions)


class ModuloCircuit:
    """
    Represents a modulo circuit capable of performing arithmetic operations on base field elements,
    storing constants, and caching powers of base field elements.

    Attributes:
        circuit_name (str): The name of the circuit.
        last_offset (int): The last used offset in the circuit's memory.
        N_LIMBS (int): The number of limbs used in the circuit.
        values_segment (dict[int, BaseFieldElement]): A dictionary mapping offsets to base field elements.
        add_offsets (list[tuple]): A list of tuples representing the offsets involved in addition operations.
        mul_offsets (list[tuple]): A list of tuples representing the offsets involved in multiplication operations.
        constants (dict[str, ModuloElement]): A dictionary mapping constant names to their ModuloElement representations.
    """

    def __init__(
        self,
        name: str,
        curve_id: int,
        generic_circuit: bool = False,
        generic_modulus: bool = False,
        compilation_mode: int = 0,
    ) -> None:
        assert (
            len(name) <= 31
        ), f"Name '{name}' is too long to fit in a felt252, size is: {len(name)}"
        self.name = name
        self.class_name = "ModuloCircuit"
        self.curve_id = curve_id
        self.field = BaseField(CURVES[curve_id].p)
        self.N_LIMBS = 4
        self.values_segment: ValueSegment = ValueSegment(
            name=name, compilation_mode=compilation_mode
        )
        self.constants: dict[int, ModuloCircuitElement] = dict()
        self.generic_circuit = generic_circuit
        self.generic_modulus = (
            generic_modulus  # Generic modulus has precedence over generic_circuit
        )
        self.compilation_mode = compilation_mode
        self.exact_output_refs_needed = None
        self.input_structs: list[Cairo1SerializableStruct] = []
        self.do_not_inline = False

    @property
    def values_offset(self) -> int:
        return self.values_segment.offset

    @property
    def output(self) -> list[ModuloCircuitElement]:
        return self.values_segment.output

    @property
    def input(self) -> list[ModuloCircuitElement]:
        return self.values_segment.input

    @property
    def output_structs(self) -> list[Cairo1SerializableStruct]:
        return self.values_segment.output_structs

    @property
    def continuous_output(self) -> bool:
        # Only useful in cairo 0 mode.
        return all(
            self.output[i + 1].offset - self.output[i].offset == N_LIMBS
            for i in range(len(self.output) - 1)
        )

    @property
    def witnesses(self) -> list[PyFelt]:
        # Only useful in cairo 0 mode.
        return [
            self.values_segment.segment_stacks[WriteOps.WITNESS][offset].felt
            for offset in sorted(self.values_segment.segment_stacks[WriteOps.WITNESS])
        ]

    def is_empty_circuit(self) -> bool:
        return len(self.values_segment.segment_stacks[WriteOps.BUILTIN]) == 0

    def write_element(
        self,
        elmt: PyFelt | int,
        write_source: WriteOps = WriteOps.INPUT,
        instruction: ModuloCircuitInstruction | None = None,
    ) -> ModuloCircuitElement:
        """
        Register an emulated field element to the circuit given its value and the write source.
        Returns a ModuloCircuitElement representing the written element with its offset as identifier.
        """
        assert isinstance(elmt, PyFelt) or isinstance(
            elmt, int
        ), f"Expected PyFelt or int, got {type(elmt)}"
        if isinstance(elmt, int):
            elmt = self.field(elmt)
        value_offset = self.values_segment.write_to_segment(
            ValueSegmentItem(
                elmt,
                write_source,
                instruction,
            )
        )
        res = ModuloCircuitElement(elmt, value_offset)
        return res

    def write_struct(
        self,
        struct: Cairo1SerializableStruct,
        write_source: WriteOps = WriteOps.INPUT,
    ) -> Union[
        ModuloCircuitElement,
        List[ModuloCircuitElement],
        List[List[Union[ModuloCircuitElement, List[ModuloCircuitElement]]]],
    ]:
        all_pyfelt = all(type(elmt) == PyFelt for elmt in struct.elmts)
        all_cairo1serializablestruct = all(
            isinstance(elmt, Cairo1SerializableStruct) for elmt in struct.elmts
        )
        assert (
            all_pyfelt or all_cairo1serializablestruct
        ), f"Expected list of PyFelt or Cairo1SerializableStruct, got {[type(elmt) for elmt in struct.elmts]}"

        if all_pyfelt:
            self.input_structs.append(struct)
            if len(struct) == 1 and isinstance(struct, (u384, u256, u128, GenericT)):
                return self.write_element(struct.elmts[0], write_source)
            else:
                return self.write_elements(struct.elmts, write_source)
        elif all_cairo1serializablestruct:
            result = [self.write_struct(elmt, write_source) for elmt in struct.elmts]
            # Ensure only the larger struct is appended
            self.input_structs = [
                s for s in self.input_structs if s not in struct.elmts
            ]
            self.input_structs.append(struct)
            return result

    def write_elements(
        self,
        elmts: list[PyFelt],
        operation: WriteOps = WriteOps.INPUT,
        sparsity: list[int] = None,
    ) -> list[ModuloCircuitElement]:
        if sparsity is not None:
            assert len(sparsity) == len(
                elmts
            ), f"Expected sparsity of length {len(elmts)}, got {len(sparsity)}"
            vals = [
                (
                    self.write_element(elmt, operation)
                    if sparsity[i] != 0
                    else self.set_or_get_constant(0)
                )
                for i, elmt in enumerate(elmts)
            ]
        else:
            vals = [self.write_element(elmt, operation, None) for elmt in elmts]
        return vals

    def write_cairo_native_felt(self, native_felt: PyFelt):
        assert (
            self.compilation_mode == 0
        ), "write_cairo_native_felt is not supported in cairo 1 mode"
        assert isinstance(
            native_felt, PyFelt
        ), f"Expected PyFelt, got {type(native_felt)}"
        assert 0 <= native_felt.value < STARK
        res = self.write_element(elmt=native_felt, write_source=WriteOps.FELT)
        return res

    def write_sparse_constant_elements(
        self, elmts: list[PyFelt]
    ) -> tuple[list[ModuloCircuitElement], list[int]]:
        sparsity = get_sparsity(elmts)
        elements = []
        for elmt, s in zip(elmts, sparsity):
            match s:
                case 0:
                    # Mocked 0 element not written to the circuit. Be careful to pass sparsity when evaluating.
                    elements.append(ModuloCircuitElement(self.field.zero(), -1))
                case 2:
                    # Mocked 1 element not written to the circuit. Be careful to pass sparsity when evaluating.
                    elements.append(ModuloCircuitElement(self.field.one(), -1))
                case _:
                    elements.append(self.set_or_get_constant(elmt.value))
        return elements, sparsity

    def set_or_get_constant(self, val: PyFelt | int) -> ModuloCircuitElement:
        if isinstance(val, int):
            val = self.field(val)
        if val.value in self.constants:
            # print((f"/!\ Constant '{hex(val.value)}' already exists."))
            return self.constants[val.value]
        self.constants[val.value] = self.write_element(val, WriteOps.CONSTANT)
        return self.constants[val.value]

    def add(
        self,
        a: ModuloCircuitElement,
        b: ModuloCircuitElement,
        comment: str | None = None,
    ) -> ModuloCircuitElement:
        if a is None and isinstance(b, ModuloCircuitElement):
            return b
        elif b is None and isinstance(a, ModuloCircuitElement):
            return a
        else:
            assert isinstance(a, ModuloCircuitElement) and isinstance(
                b, ModuloCircuitElement
            ), f"Expected ModuloElement, got {type(a)}, {a} and {type(b)}, {b}"

            instruction = ModuloCircuitInstruction(
                ModBuiltinOps.ADD, a.offset, b.offset, self.values_offset, comment
            )
            return self.write_element(
                a.emulated_felt + b.emulated_felt, WriteOps.BUILTIN, instruction
            )

    def sum(self, args: list[ModuloCircuitElement], comment: str | None = None):
        if not args:
            raise ValueError("The 'args' list cannot be empty.")
        assert all(isinstance(elmt, ModuloCircuitElement) for elmt in args)
        result = args[0]
        for elmt in args[1:]:
            result = self.add(result, elmt, comment)
        return result

    def product(self, args: list[ModuloCircuitElement], comment: str | None = None):
        if not args:
            raise ValueError("The 'args' list cannot be empty.")
        assert all(isinstance(elmt, ModuloCircuitElement) for elmt in args)
        result = args[0]
        for elmt in args[1:]:
            result = self.mul(result, elmt, comment)
        return result

    def double(self, a: ModuloCircuitElement) -> ModuloCircuitElement:
        return self.add(a, a)

    def mul(
        self,
        a: ModuloCircuitElement,
        b: ModuloCircuitElement,
        comment: str | None = None,
    ) -> ModuloCircuitElement:
        if a is None and isinstance(b, ModuloCircuitElement):
            return self.set_or_get_constant(0)
        elif b is None and isinstance(a, ModuloCircuitElement):
            return self.set_or_get_constant(0)
        assert isinstance(a, ModuloCircuitElement) and isinstance(
            b, ModuloCircuitElement
        ), f"Expected ModuloElement, got lhs {type(a)}, {a} and rhs {type(b)}, {b}"
        instruction = ModuloCircuitInstruction(
            ModBuiltinOps.MUL, a.offset, b.offset, self.values_offset, comment
        )
        return self.write_element(
            a.emulated_felt * b.emulated_felt, WriteOps.BUILTIN, instruction
        )

    def square(
        self, a: ModuloCircuitElement, comment: str | None = None
    ) -> ModuloCircuitElement:
        return self.mul(a, a, comment)

    def neg(
        self, a: ModuloCircuitElement, comment: str | None = None
    ) -> ModuloCircuitElement:
        assert (
            type(a) == ModuloCircuitElement
        ), f"Expected ModuloElement, got {type(a)}, {a}"
        res = self.sub(self.set_or_get_constant(self.field.zero()), a, comment)
        return res

    def sub(
        self,
        a: ModuloCircuitElement,
        b: ModuloCircuitElement,
        comment: str | None = None,
    ):
        assert isinstance(a, ModuloCircuitElement) and isinstance(
            b, ModuloCircuitElement
        ), f"Expected ModuloElement, got {type(a)}, {a} and {type(b)}, {b}"
        instruction = ModuloCircuitInstruction(
            ModBuiltinOps.ADD, b.offset, self.values_offset, a.offset, comment
        )
        return self.write_element(a.felt - b.felt, WriteOps.BUILTIN, instruction)

    def inv(
        self,
        a: ModuloCircuitElement,
        comment: str | None = None,
    ):
        assert isinstance(
            a, ModuloCircuitElement
        ), f"Expected ModuloElement, got {type(a)}, {a}"
        if self.compilation_mode == 0:
            one = self.set_or_get_constant(
                1
            )  # Write one before accessing its offset so self.values_offset is correctly updated.

            instruction = ModuloCircuitInstruction(
                ModBuiltinOps.MUL,
                a.offset,
                self.values_offset,
                one.offset,
                comment,
            )
        elif self.compilation_mode == 1:
            instruction = ModuloCircuitInstruction(
                ModBuiltinOps.MUL,
                a.offset,
                None,
                None,
                comment,
            )

        return self.write_element(a.felt.__inv__(), WriteOps.BUILTIN, instruction)

    def div(
        self,
        a: ModuloCircuitElement,
        b: ModuloCircuitElement,
        comment: str | None = None,
    ):
        assert isinstance(a, ModuloCircuitElement) and isinstance(
            b, ModuloCircuitElement
        ), f"Expected ModuloElement, got {type(a)}, {a} and {type(b)}, {b}"
        if self.compilation_mode == 0:
            instruction = ModuloCircuitInstruction(
                ModBuiltinOps.MUL, b.offset, self.values_offset, a.offset, comment
            )
            return self.write_element(
                a.felt * b.felt.__inv__(), WriteOps.BUILTIN, instruction
            )
        else:
            return self.mul(a, self.inv(b))

    def fp_sqrt(self, element: ModuloCircuitElement) -> ModuloCircuitElement:
        """
        Computes "a" square root of a field element.
        /!\ Warning : This circuit is non deterministic /!\
        /!\ Two square roots exist for any non-zero element, and no constraint is enforced to select any of them /!\
        Raises ValueError if the element is not a quadratic residue.
        """
        assert self.compilation_mode == 0, "fp_sqrt is not supported in cairo 1 mode"

        root = element.felt.sqrt(min_root=False)

        # Write the root as a witness and verify it
        root = self.write_element(root, WriteOps.WITNESS)
        self.mul_and_assert(root, root, element, comment="Fp sqrt")
        return root

    def fp_is_non_zero(self, a: ModuloCircuitElement) -> ModuloCircuitElement:
        """
        Returns 1 if a != 0, 0 if a == 0.
        Enforces soundness by forcing `flag` to be a true indicator of whether 'a' is non-zero.

        Main Constraints:
        1) flag * (1 - flag) = 0
           - Ensures that `flag` ∈ {0,1}, i.e., it's a binary indicator.
        2) Prover supplies a_inv as a “would-be” inverse of `a`.
           - If a != 0, a_inv should be the genuine inverse, i.e. a * a_inv = 1.
           - If a == 0, a_inv is irrelevant and effectively ignored.
        3) flag*(a * a_inv) + (1 - flag)*(a + 1) == 1
           - If a == 0 → the only way is flag=0, so:
                 0 * (0 * a_inv) + 1 * (0 + 1) = 1
             which is satisfied.
           - If a != 0 → the only way is flag=1, so:
                 1 * (a * a_inv) + 0 * (a + 1) = 1
             forcing a_inv to be a true inverse of a.

        Consequently:
          • When a == 0, the circuit forces flag=0 and ignores a_inv.
          • When a != 0, the circuit forces flag=1 and requires a_inv to be the correct inverse.

        Overall, this design prevents cheating (e.g. claiming a != 0 when it's actually 0, or supplying a bogus inverse) because the single equation in step (3) couples a, a_inv, and flag in a way that only the correct assignments can satisfy all constraints.
        """

        assert (
            self.compilation_mode == 0
        ), "`fp_is_non_zero_flag` is only supported in cairo 0 mode."

        # Step 1: Prover declares a boolean-labeled witness: 1 if a != 0, else 0
        is_non_zero_python_bool = int(a.value != 0)
        flag = self.write_element(is_non_zero_python_bool, WriteOps.WITNESS)

        # Step 2: Force the flag to be a single bit: flag * (1 - flag) == 0

        one_minus_flag = self.sub(self.set_or_get_constant(1), flag)
        self.mul_and_assert(
            flag,
            one_minus_flag,
            self.set_or_get_constant(0),
            comment="flag ∈ {0,1}",
        )

        # Step 3: The prover supplies a_inv as a witness.
        #          If flag=1 (a != 0), a_inv must be the real inverse: a*a_inv = 1.
        #          If flag=0 (a == 0), a_inv is irrelevant.
        a_inv_val = 0
        if is_non_zero_python_bool:
            a_inv_val = a.felt.__inv__().value
        a_inv = self.write_element(a_inv_val, WriteOps.WITNESS)

        # Step 4: Combine them with the core check:
        #         flag*(a*a_inv) + (1 - flag)*(a+1) == 1
        #         => forces correct assignments for (flag, a_inv)
        flag_eq_0_branch = self.add(a, self.set_or_get_constant(1))
        flag_eq_1_branch = self.mul(a, a_inv)

        self.add_and_assert(
            self.mul(flag, flag_eq_1_branch),
            self.mul(one_minus_flag, flag_eq_0_branch),
            self.set_or_get_constant(1),
        )

        # Return the final flag: 1 if a != 0, 0 if a == 0
        return flag

    def vector_sub(
        self, X: list[ModuloCircuitElement], Y: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        return [
            self.sub(x, y, comment=f"Fp{len(X)} sub coeff {i}/{len(X)-1}")
            for i, (x, y) in enumerate(zip(X, Y))
        ]

    def vector_scale(
        self, X: list[ModuloCircuitElement], c: ModuloCircuitElement
    ) -> list[ModuloCircuitElement]:
        """
        Multiplies a polynomial with coefficients `X` by a scalar `c`.
        Input : I(x) = i0 + i1*x + i2*x^2 + ... + in-1*x^n-1
        Output : O(x) = ci0 + ci1*x + ci2*x^2 + ... + cin-1*x^n-1.
        This is done in the circuit.
        """
        assert isinstance(c, ModuloCircuitElement), "c must be a ModuloCircuitElement"
        return [
            self.mul(x_i, c, comment=f"Fp{len(X)} scalar mul coeff {i}/{len(X)-1}")
            for i, x_i in enumerate(X)
        ]

    def vector_add(
        self, X: list[ModuloCircuitElement], Y: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        """
        Adds two polynomials with coefficients `X` and `Y`.
        Returns R = [x0 + y0, x1 + y1, x2 + y2, ... + xn-1 + yn-1] mod p
        """
        assert len(X) == len(Y), f"len(X)={len(X)} != len(Y)={len(Y)}"
        return [
            self.add(x_i, y_i, comment=f"Fp{len(X)} add coeff {i}/{len(X)-1}")
            for i, (x_i, y_i) in enumerate(zip(X, Y))
        ]

    def vector_neg(self, X: list[ModuloCircuitElement]) -> list[ModuloCircuitElement]:
        """
        Negates a polynomial with coefficients `X`.
        Returns R = [-x0, -x1, -x2, ... -xn-1] mod p
        """
        return [
            self.neg(x_i, comment=f"Fp{len(X)} neg coeff {i}/{len(X)-1}")
            for i, x_i in enumerate(X)
        ]

    def sub_and_assert(
        self,
        a: ModuloCircuitElement,
        b: ModuloCircuitElement,
        c: ModuloCircuitElement,
        comment: str | None = None,
    ):
        """
        Subtracts b from a and asserts that the result is equal to c.
        In practice, it checks that c + b = a [mod p].
        All three values are expected to be already in the value segment, no new value is created.
        Costs 2 Steps.
        """
        assert (
            self.compilation_mode == 0
        ), "sub_and_assert is not supported in cairo 1 mode"
        instruction = ModuloCircuitInstruction(
            ModBuiltinOps.ADD, c.offset, b.offset, a.offset, comment
        )
        self.values_segment.assert_eq_instructions.append(instruction)
        return c

    def add_and_assert(
        self,
        a: ModuloCircuitElement,
        b: ModuloCircuitElement,
        c: ModuloCircuitElement,
        comment: str | None = None,
    ):
        """
        Adds a and b and asserts that the result is equal to c.
        In practice, it only checks that a + b = c [mod p].
        All three values are expected to be already in the value segment, no new value is created.
        Costs 2 Steps.
        """
        assert (
            self.compilation_mode == 0
        ), "add_and_assert is not supported in cairo 1 mode"
        instruction = ModuloCircuitInstruction(
            ModBuiltinOps.ADD, a.offset, b.offset, c.offset, comment
        )
        self.values_segment.assert_eq_instructions.append(instruction)
        return c

    def mul_and_assert(
        self,
        a: ModuloCircuitElement,
        b: ModuloCircuitElement,
        c: ModuloCircuitElement,
        comment: str | None = None,
    ):
        assert (
            self.compilation_mode == 0
        ), "mul_and_assert is not supported in cairo 1 mode"

        instruction = ModuloCircuitInstruction(
            ModBuiltinOps.MUL, a.offset, b.offset, c.offset, comment
        )
        self.values_segment.assert_eq_instructions.append(instruction)
        return c

    def eval_horner(
        self,
        poly: list[ModuloCircuitElement],
        z: ModuloCircuitElement,
        poly_name: str = None,
        var_name: str = "z",
    ):
        """
        Evaluates a polynomial at point z using Horner's method.
        Assumes that the polynomial is in the form a0 + a1*z + a2*z^2 + ... + an*z^n, indexed with the constant coefficient first.
        """
        if poly_name is None:
            poly_name = "UnnamedPoly"

        # Regular Horner evaluation
        acc = poly[-1]  # Start with the highest degree coefficient
        for i in range(len(poly) - 2, -1, -1):
            acc = self.add(
                poly[i],
                self.mul(
                    acc,
                    z,
                    comment=f"Eval {poly_name} Horner step: multiply by {var_name}",
                ),
                comment=f"Eval {poly_name} Horner step: add coefficient_{i}",
            )

        return acc

    def eval_poly(
        self,
        poly: list[ModuloCircuitElement],
        X_powers: list[ModuloCircuitElement],
        poly_name: str = None,
        var_name: str = "x",
    ):
        """
        Evaluates a polynomial at precomputed powers of X.
        Assumes that the polynomial is in the form a0 + a1*x + a2*x^2 + ... + an*x^n, indexed with the constant coefficient first.
        Assumes that X_powers is a list of powers of X, such that X_powers[i] = X^(i+1).
        """
        if poly_name is None:
            poly_name = "UnnamedPoly"
        assert len(poly) - 1 <= len(
            X_powers
        ), f"Expected at least {len(poly) - 1} powers of X to evaluate P, got {len(X_powers)}"

        acc = poly[0]
        for i in range(1, len(poly)):
            acc = self.add(
                acc,
                self.mul(
                    poly[i],
                    X_powers[i - 1],
                    comment=f"Eval {poly_name} step coeff_{i} * {var_name}^{i}",
                ),
                comment=f"Eval {poly_name} step + (coeff_{i} * {var_name}^{i})",
            )
        return acc

    def extend_output(self, elmts: list[ModuloCircuitElement]):
        """
        Adds elements to the output of the circuit.
        """
        assert isinstance(elmts, (list, tuple))
        assert all(isinstance(x, ModuloCircuitElement) for x in elmts)
        self.output.extend(elmts)
        return

    def extend_struct_output(self, struct: Cairo1SerializableStruct):
        """
        Adds elements to the output of the circuit in struct form.
        """
        assert isinstance(struct, Cairo1SerializableStruct)
        if self.compilation_mode == 0:
            self.extend_output(struct.elmts)
        else:
            self.values_segment.output_structs.append(struct)
            self.extend_output(struct.elmts)
        return

    def print_value_segment(self):
        self.values_segment.print()

    def compile_circuit(
        self,
        function_name: str = None,
        pub: bool = True,
        inline: bool = True,
        generic_modulus: bool = False,
    ):
        if self.is_empty_circuit():
            return "", ""
        self.values_segment = self.values_segment.non_interactive_transform()
        if self.compilation_mode == 0:
            return self.compile_circuit_cairo_zero(function_name), None
        elif self.compilation_mode == 1:
            return self.compile_circuit_cairo_1(
                function_name, pub, inline, generic_modulus
            )

    def compile_circuit_cairo_zero(
        self,
        function_name: str = None,
        returns: dict[str] = {
            "felt*": [
                "constants_ptr",
                "add_offsets_ptr",
                "mul_offsets_ptr",
                "output_offsets_ptr",
            ],
            "felt": [
                "constants_ptr_len",
                "input_len",
                "witnesses_len",
                "output_len",
                "continuous_output",
                "add_mod_n",
                "mul_mod_n",
                "n_assert_eq",
                "name",
                "curve_id",
            ],
        },
    ) -> str:
        dw_arrays = self.values_segment.get_dw_lookups()
        name = function_name or self.values_segment.name
        function_name = f"get_{name}_circuit"
        if self.generic_circuit:
            code = (
                f"func {function_name}(curve_id:felt)->(circuit:{self.class_name}*)"
                + "{"
                + "\n"
            )
        else:
            code = f"func {function_name}()->(circuit:{self.class_name}*)" + "{" + "\n"

        code += "alloc_locals;\n"
        code += "let (__fp__, _) = get_fp_and_pc();\n"

        for dw_array_name in returns["felt*"]:
            code += f"let ({dw_array_name}:felt*) = get_label_location({dw_array_name}_loc);\n"

        code += f"let constants_ptr_len = {len(dw_arrays['constants_ptr'])};\n"
        code += f"let input_len = {len(self.values_segment.segment_stacks[WriteOps.INPUT])*N_LIMBS};\n"
        code += f"let witnesses_len = {len(self.values_segment.segment_stacks[WriteOps.WITNESS])*N_LIMBS};\n"
        code += f"let output_len = {len(self.output)*N_LIMBS};\n"
        continuous_output = self.continuous_output
        code += f"let continuous_output = {1 if continuous_output else 0};\n"
        code += f"let add_mod_n = {len(dw_arrays['add_offsets_ptr'])};\n"
        code += f"let mul_mod_n = {len(dw_arrays['mul_offsets_ptr'])};\n"
        code += (
            f"let n_assert_eq = {len(self.values_segment.assert_eq_instructions)};\n"
        )
        code += f"let name = '{self.name}';\n"
        code += (
            f"let curve_id = {'curve_id' if self.generic_circuit else self.curve_id};\n"
        )

        code += f"local circuit:{self.class_name} = {self.class_name}({', '.join(returns['felt*'])}, {', '.join(returns['felt'])});\n"
        code += "return (&circuit,);\n"

        for dw_array_name in returns["felt*"]:
            dw_values = dw_arrays[dw_array_name]
            code += f"\t {dw_array_name}_loc:\n"
            if dw_array_name == "constants_ptr":
                for bigint in dw_values:
                    for limb in bigint:
                        code += f"\t dw {limb};\n"
                code += "\n"

            elif dw_array_name in ["add_offsets_ptr", "mul_offsets_ptr"]:
                num_instructions = len(dw_values)
                instructions_needed = (
                    BATCH_SIZE - (num_instructions % BATCH_SIZE)
                ) % BATCH_SIZE  # Must be a multiple of 8 (currently)
                for left, right, result, comment in dw_values:
                    code += (
                        f"\t dw {left}; // {comment}\n"
                        + f"\t dw {right};\n"
                        + f"\t dw {result};\n"
                    )
                if instructions_needed > 0:
                    first_triplet = dw_values[0]
                    for _ in range(instructions_needed):
                        code += (
                            f"\t dw {first_triplet[0]};\n"
                            + f"\t dw {first_triplet[1]};\n"
                            + f"\t dw {first_triplet[2]};\n"
                        )
                code += "\n"
            elif dw_array_name in ["output_offsets_ptr"]:
                if continuous_output:
                    code += f"\t dw {dw_values[0]};\n"
                else:
                    for val in dw_values:
                        code += f"\t dw {val};\n"

        code += "\n"
        code += "}\n"
        return code

    def write_cairo1_input_stack(
        self,
        write_ops: WriteOps,
        code: str,
        offset_to_reference_map: dict[int, str],
        start_index: int,
    ) -> tuple:
        """
        Defines the inputs for the compiled Cairo 1 circuit.
        """
        len_stack = len(self.values_segment.segment_stacks[write_ops])
        if len_stack > 0:
            code += f"\n // {write_ops.name} stack\n"
            offsets = list(self.values_segment.segment_stacks[write_ops].keys())
            i = 0
            while i < len_stack:
                if write_ops == WriteOps.CONSTANT:
                    val = self.values_segment.segment[offsets[i]].value
                    if 0 < -val % self.field.p <= 100:
                        comment = f"// -{hex(-val%self.field.p)} % p"
                    else:
                        comment = f"// {hex(val)}"
                    code += f"\t let in{start_index+i} = CE::<CI<{start_index+i}>> {{}}; {comment}\n"
                    offset_to_reference_map[offsets[i]] = f"in{start_index+i}"
                    i += 1
                else:
                    if i + 2 < len_stack:
                        code += f"\t let (in{start_index+i}, in{start_index+i+1}, in{start_index+i+2}) = (CE::<CI<{start_index+i}>> {{}}, CE::<CI<{start_index+i+1}>> {{}}, CE::<CI<{start_index+i+2}>> {{}});\n"
                        offset_to_reference_map[offsets[i]] = f"in{start_index+i}"
                        offset_to_reference_map[offsets[i + 1]] = f"in{start_index+i+1}"
                        offset_to_reference_map[offsets[i + 2]] = f"in{start_index+i+2}"
                        i += 3
                    elif i + 1 < len_stack:
                        code += f"\t let (in{start_index+i}, in{start_index+i+1}) = (CE::<CI<{start_index+i}>> {{}}, CE::<CI<{start_index+i+1}>> {{}});\n"
                        offset_to_reference_map[offsets[i]] = f"in{start_index+i}"
                        offset_to_reference_map[offsets[i + 1]] = f"in{start_index+i+1}"
                        i += 2
                    else:
                        code += f"\t let in{start_index+i} = CE::<CI<{start_index+i}>> {{}};\n"
                        offset_to_reference_map[offsets[i]] = f"in{start_index+i}"
                        i += 1
            return (
                code,
                offset_to_reference_map,
                start_index + len_stack,
            )
        else:
            return code, offset_to_reference_map, start_index

    def fill_cairo_1_constants(self) -> tuple[str, str]:
        """
        Return the part of the code that fills the constants to the circuit, and the constants array
        if there are more than 8 constants. In that case, .next_span() is used to save bytecode and
        the constants are stored in a const array.
        """
        constants_ints = [
            self.values_segment.segment[offset].value
            for offset in self.values_segment.segment_stacks[WriteOps.CONSTANT].keys()
        ]
        if len(constants_ints) < 8:
            constants_split = [bigint_split(x, N_LIMBS, BASE) for x in constants_ints]
            constants_filled = "\n".join(
                f"circuit_inputs = circuit_inputs.next_2([{','.join(hex(x) for x in constants_split[i])}]); // in{i}"
                for i in range(len(constants_ints))
            )
            return constants_filled, None
        else:
            import garaga.hints.io as io

            const_name = (
                self.name.upper()
                + "_"
                + CurveID(self.curve_id).name.upper()
                + "_CONSTANTS"
            )
            constants_filled = f"""
            circuit_inputs = circuit_inputs.next_span({const_name}.span()); // in{0} - in{len(constants_ints)-1}
            """

            const_array = f"const {const_name}: [u384; {len(constants_ints)}] = {io.int_array_to_u384_array(constants_ints, const=True)};"
        return constants_filled, const_array

    def write_cairo1_circuit(self, offset_to_reference_map: dict[int, str]) -> str:
        """
        Defines the arithmetic instructions for the compiled Cairo 1 circuit.
        """
        code = ""
        for i, (offset, vs_item) in enumerate(
            self.values_segment.segment_stacks[WriteOps.BUILTIN].items()
        ):
            op = vs_item.instruction.operation
            left_offset = vs_item.instruction.left_offset
            right_offset = vs_item.instruction.right_offset
            result_offset = vs_item.instruction.result_offset
            comment = vs_item.instruction.comment or ""

            # print(i, op, left_offset, right_offset, result_offset)
            match op:
                case ModBuiltinOps.ADD:
                    if right_offset > result_offset:
                        # Case sub
                        code += f"let t{i} = circuit_sub({offset_to_reference_map[result_offset]}, {offset_to_reference_map[left_offset]}); {'// '+comment if comment else ''}\n"
                        offset_to_reference_map[offset] = f"t{i}"
                        assert offset == right_offset
                    else:
                        code += f"let t{i} = circuit_add({offset_to_reference_map[left_offset]}, {offset_to_reference_map[right_offset]}); {'// '+comment if comment else ''}\n"
                        offset_to_reference_map[offset] = f"t{i}"
                        assert offset == result_offset

                case ModBuiltinOps.MUL:
                    if right_offset == result_offset == offset:
                        # Case inv
                        # print(f"\t INV {left_offset} {right_offset} {result_offset}")
                        code += f"let t{i} = circuit_inverse({offset_to_reference_map[left_offset]}); {'// '+comment if comment else ''}\n"
                        offset_to_reference_map[offset] = f"t{i}"
                    else:
                        # print(f"MUL {left_offset} {right_offset} {result_offset}")
                        code += f"let t{i} = circuit_mul({offset_to_reference_map[left_offset]}, {offset_to_reference_map[right_offset]}); {'// '+comment if comment else ''}\n"
                        offset_to_reference_map[offset] = f"t{i}"
                        assert offset == result_offset
        return code

    def compile_circuit_cairo_1(
        self,
        function_name: str = None,
        pub: bool = False,
        inline: bool = True,
        generic_modulus: bool = False,
    ) -> tuple[str, str]:
        """
        Defines the Cairo 1 function code for the compiled circuit.
        """
        name = function_name or self.values_segment.name
        function_name = f"run_{name}_circuit"
        curve_index = CurveID.find_value_in_string(name)
        return_is_struct = False
        input_is_struct = False
        if self.output_structs:
            if sum([len(x.elmts) for x in self.output_structs]) == len(self.output):
                return_is_struct = True
                signature_output = (
                    f"({','.join([x.struct_name for x in self.output_structs])}"
                )

                signature_output += ",)" if len(self.output_structs) == 1 else ")"
            else:
                raise ValueError(
                    f"Output structs must have the same number of elements as the output: {len(self.output_structs)=} != {len(self.output)=}"
                )
        else:
            signature_output = "Array<u384>"

        if self.input_structs:
            if sum([len(x) for x in self.input_structs]) == len(self.input):
                input_is_struct = True
                signature_input = f"{','.join([x.serialize_input_signature() for x in self.input_structs])}"
            else:
                raise ValueError(
                    f"Input structs must have the same number of elements as the input: {sum([len(x) for x in self.input_structs])=} != {len(self.input)=}"
                )
        else:
            signature_input = "mut input: Array<u384>"
        if "<T>" in signature_input or ":T," in signature_input:
            generic_input = "<T, +IntoCircuitInputValue<T>, +Drop<T>, +Copy<T>>"
        else:
            generic_input = ""
        if pub:
            prefix = "pub "
        else:
            prefix = ""
        if inline:
            if generic_input == "":
                attribute = "#[inline(always)]"
            else:
                attribute = "#[inline]"
        else:
            attribute = ""
        generic_modulus = generic_modulus or self.generic_modulus
        if generic_modulus:
            code = f"{attribute}\n{prefix}fn {function_name}{generic_input}({signature_input}, modulus:CircuitModulus)->{signature_output} {{\n"
        elif self.generic_circuit:
            code = f"{attribute}\n{prefix}fn {function_name}{generic_input}({signature_input}, curve_index:usize)->{signature_output} {{\n"
        else:
            code = f"{attribute}\n{prefix}fn {function_name}{generic_input}({signature_input})->{signature_output} {{\n"

        # Define the input for the circuit.
        code, offset_to_reference_map, start_index = self.write_cairo1_input_stack(
            WriteOps.CONSTANT, code, {}, 0
        )
        (
            code,
            offset_to_reference_map,
            commit_start_index,
        ) = self.write_cairo1_input_stack(
            WriteOps.INPUT, code, offset_to_reference_map, start_index
        )
        code, offset_to_reference_map, commit_end_index = self.write_cairo1_input_stack(
            WriteOps.COMMIT, code, offset_to_reference_map, commit_start_index
        )
        code, offset_to_reference_map, start_index = self.write_cairo1_input_stack(
            WriteOps.WITNESS, code, offset_to_reference_map, commit_end_index
        )
        code, offset_to_reference_map, start_index = self.write_cairo1_input_stack(
            WriteOps.FELT, code, offset_to_reference_map, start_index
        )

        code += self.write_cairo1_circuit(offset_to_reference_map)

        outputs_refs = []
        for out in self.output:
            if self.values_segment[out.offset].write_source == WriteOps.BUILTIN:
                outputs_refs.append(offset_to_reference_map[out.offset])
            else:
                continue
        if self.exact_output_refs_needed:
            outputs_refs_needed = [
                offset_to_reference_map[out.offset]
                for out in self.exact_output_refs_needed
            ]
        else:
            outputs_refs_needed = outputs_refs

        if generic_modulus:
            code += "\nlet modulus = modulus;\n"
        elif curve_index is not None:
            code += f"""
    let modulus = get_{CurveID(self.curve_id).name}_modulus(); // {CurveID(self.curve_id).name} prime field modulus
        """
        else:
            code += """
    let modulus = get_modulus(curve_index);
        """

        code += f"""
    let mut circuit_inputs = ({','.join(outputs_refs_needed)},).new_inputs();
    // Prefill constants:
    """

        tmp, const_array = self.fill_cairo_1_constants()
        code += tmp
        code += """
        // Fill inputs:
        """

        acc_len = len(self.values_segment.segment_stacks[WriteOps.CONSTANT])
        if input_is_struct:
            for struct in self.input_structs:
                struct_code = struct.dump_to_circuit_input()
                struct_code_lines = [
                    line for line in struct_code.split("\n") if line.strip()
                ]
                if "while let" not in struct_code and "for val in" not in struct_code:
                    struct_code_with_counter = "\n".join(
                        f"{line} // in{i+acc_len}"
                        for i, line in enumerate(struct_code_lines)
                    )
                else:
                    struct_code_with_counter = (
                        struct_code + f" // in{acc_len} - in{acc_len+len(struct)-1}\n"
                    )
                acc_len += len(struct)
                code += struct_code_with_counter + "\n"
        else:
            code += """
    let mut input = input;
    while let Option::Some(val) = input.pop_front() {
        circuit_inputs = circuit_inputs.next(val);
    };
    """
        code += """
        let outputs = circuit_inputs.done_2().eval(modulus).unwrap();
"""
        if return_is_struct:
            code += "\n".join(
                [
                    struct.extract_from_circuit_output(offset_to_reference_map)
                    for struct in self.output_structs
                ]
            )
            code += (
                f"return ({','.join([struct.name for struct in self.output_structs])}"
            )
            code += ",);\n}" if len(self.output_structs) == 1 else ");\n}"
        else:
            code += f"let res=array![{','.join([f'outputs.get_output({ref})' for ref in outputs_refs])}];\n"
            code += "return res;\n"
            code += "}\n"

        if const_array:
            # Add the constants outside of the function if they are more than 8.
            code += "\n"
            code += const_array
        return code, function_name

    def summarize(self):
        add_count, mul_count, assert_eq_count = self.values_segment.summarize()
        summary = {
            "circuit": self.name,
            "MULMOD": mul_count,
            "ADDMOD": add_count,
            "ASSERT_EQ": assert_eq_count,
            "POSEIDON": 0,
            "RLC": 0,
        }

        return summary


if __name__ == "__main__":
    from garaga.algebra import BaseField

    field = BaseField(256)
    circuit = ModuloCircuit("test_circuit", 0)
    a = circuit.write_element(field(1))
    b = circuit.write_element(field(2))
    c = circuit.add(a, b)
    d = circuit.mul(c, b)

    print(c)

    print(d)

    circuit = ModuloCircuit("test_empty_circuit", curve_id=0, compilation_mode=1)
    print(circuit.compile_circuit())



import argparse
import os
import subprocess
from pathlib import Path

from garaga.definitions import CurveID, ProofSystem
from garaga.modulo_circuit_structs import G2Line, StructArray
from garaga.precompiled_circuits.compilable_circuits.base import (
    get_circuit_definition_impl_template,
)
from garaga.precompiled_circuits.compilable_circuits.ultra_honk import (
    PrepareScalarsCircuit,
    SumCheckCircuit,
    ZKEvalsConsistencyDoneCircuit,
    ZKEvalsConsistencyInitCircuit,
    ZKEvalsConsistencyLoopCircuit,
    ZKPrepareScalarsCircuit,
    ZKSumCheckCircuit,
)
from garaga.precompiled_circuits.honk import (
    CONST_PROOF_SIZE_LOG_N,
    G2_POINT_KZG_1,
    G2_POINT_KZG_2,
    MAX_LOG_N,
    PAIRING_POINT_OBJECT_LENGTH,
    HonkVk,
)
from garaga.precompiled_circuits.multi_miller_loop import precompute_lines
from garaga.starknet.cli.utils import create_directory
from garaga.starknet.groth16_contract_generator.generator import (
    CAIRO_VERSION,
    get_scarb_toml_file,
)

# nargo --version
# bb --version

BB_VERSION = "0.86.0"
BBUP_VERSION = "0.86.0-starknet.1"
NARGO_VERSION = "1.0.0-beta.3"


def gen_honk_verifier(
    vk: str | Path | HonkVk | bytes,
    output_folder_path: str,
    output_folder_name: str,
    system: ProofSystem = ProofSystem.UltraKeccakZKHonk,
    cli_mode: bool = False,
) -> bool:
    """
    Generate the honk verifier files and write them to the output folder.
    """
    output_folder_path = os.path.join(output_folder_path, output_folder_name)

    constants_code, circuits_code, contract_code = gen_honk_verifier_files(vk, system)
    _write_and_format_project_files(
        output_folder_path,
        output_folder_name,
        cli_mode,
        constants_code,
        circuits_code,
        contract_code,
    )
    return True


def gen_honk_verifier_files(
    vk: str | Path | HonkVk | bytes,
    system: ProofSystem = ProofSystem.UltraKeccakZKHonk,
) -> tuple[str, str, str]:
    """
    Returns in this order:
    - constants_code
    - circuits_code
    - contract_code
    """
    match system:
        case ProofSystem.UltraKeccakHonk | ProofSystem.UltraStarknetHonk:
            return _gen_honk_verifier_files(vk, system)
        case ProofSystem.UltraKeccakZKHonk | ProofSystem.UltraStarknetZKHonk:
            return _gen_zk_honk_verifier_files(vk, system)
        case _:
            raise ValueError(f"Proof system {system} not compatible")


def get_msm_kzg_template(
    msm_size: int,
    n_vk_points: int,
    n_proof_points: int,
    is_on_curve_function_name: str,
):
    TEMPLATE = """\n
            // Check input points are on curve.
            // Skip the first {n_vk_points} points as they are from VK and keep the last {n_proof_points} proof points
            for point in points.slice({n_vk_points}, {n_proof_points}) {{
                assert({is_on_curve_function_name}(*point, mod_bn), 'proof point not on curve');
            }};

            // Assert shplonk_q is on curve
            let shplonk_q_pt:G1Point = full_proof.proof.shplonk_q.into();
            assert({is_on_curve_function_name}(shplonk_q_pt, mod_bn), 'shplonk_q not on curve');

            let mut msm_hint = full_proof.msm_hint;
            assert(msm_hint.len() == {msm_len} * 12, 'wrong glv&fakeglv hint size');
            let eigen = get_eigenvalue(0);
            let third_root_of_unity = get_third_root_of_unity(0);
            let min_one = get_min_one_order(0);
            let nG = get_nG_glv_fake_glv(0);

            let mut P_1 = shplonk_q_pt;

            while msm_hint.len() != 0 {{
                let pt = *points.pop_front().unwrap();
                let scalar = *scalars.pop_front().unwrap();
                // Note : Scalars are below curve order by construction (circuit outputs mod n and transcript output (u128))
                let glv_fake_glv_hint: GlvFakeGlvHint = Serde::deserialize(ref msm_hint).unwrap();
                let temp = _scalar_mul_glv_and_fake_glv(
                    pt,
                    scalar,
                    mod_grumpkin,
                    mod_bn,
                    glv_fake_glv_hint,
                    eigen,
                    third_root_of_unity,
                    min_one,
                    nG,
                    0,
                );
                P_1 = _ec_safe_add(P_1, temp, mod_bn, 0);
            }}

            let P_2:G1Point = full_proof.proof.kzg_quotient.into();

            // Perform the KZG pairing check.
            let kzg_check = multi_pairing_check_bn254_2P_2F(
               G1G2Pair {{ p: P_1, q: G2_POINT_KZG_1 }},
               G1G2Pair {{ p: P_2.negate(0), q: G2_POINT_KZG_2 }},
               precomputed_lines.span(),
               full_proof.kzg_hint,
            );

        """.format(
        msm_len=msm_size,
        n_vk_points=n_vk_points,
        n_proof_points=n_proof_points,
        is_on_curve_function_name=is_on_curve_function_name,
    )
    return TEMPLATE


def setup_vk_flavor(system: ProofSystem, vk: str | Path | HonkVk | bytes):
    match system:
        case ProofSystem.UltraKeccakHonk | ProofSystem.UltraKeccakZKHonk:
            flavor = "Keccak"
        case ProofSystem.UltraStarknetHonk | ProofSystem.UltraStarknetZKHonk:
            flavor = "Starknet"
        case _:
            raise ValueError(f"Proof system {system} not compatible")
    if isinstance(vk, (Path, str)):
        vk = HonkVk.from_bytes(open(vk, "rb").read())
    elif isinstance(vk, bytes):
        vk = HonkVk.from_bytes(vk)
    else:
        assert isinstance(
            vk, HonkVk
        ), f"Invalid type for vk: {type(vk)}. Should be str, Path, HonkVk or bytes."
    return vk, flavor


def _get_circuit_code_header():
    header = """
use core::circuit::{
    u384, circuit_add, circuit_sub, circuit_mul, circuit_inverse,
    EvalCircuitTrait, CircuitOutputsTrait, CircuitInputs, CircuitModulus,
};
use garaga::core::circuit::{AddInputResultTrait2, u288IntoCircuitInputValue, IntoCircuitInputValue};
use core::circuit::CircuitElement as CE;
use core::circuit::CircuitInput as CI;
use garaga::definitions::{G1Point};\n
"""
    return header


def _gen_circuits_code(
    vk: HonkVk, is_zk: bool
) -> tuple[str, str, str, list[str], str, int, list[str]]:
    """
    Generate the code for the sumcheck circuit and related circuits.

    Args:
        vk: The verifying key
        is_zk: Whether to generate ZK-specific circuits

    Returns:
        Tuple containing:
        - The generated code
        - Sumcheck function name
        - Prepare scalars function name
        - List of consistency function names (empty for non-ZK)
        - Scalar indexes
        - MSM length
    """
    code = _get_circuit_code_header()
    curve_id = CurveID.GRUMPKIN

    # Generate sumcheck circuit
    sumcheck_circuit = ZKSumCheckCircuit(vk) if is_zk else SumCheckCircuit(vk)
    sumcheck_function_name = f"{curve_id.name}_{sumcheck_circuit.name.upper()}"
    sumcheck_code, sumcheck_function_name = sumcheck_circuit.circuit.compile_circuit(
        function_name=sumcheck_function_name, pub=True, generic_modulus=True
    )

    # Generate prepare scalars circuit
    prepare_scalars_circuit = (
        ZKPrepareScalarsCircuit(vk) if is_zk else PrepareScalarsCircuit(vk)
    )
    scalar_indexes = prepare_scalars_circuit.scalar_indexes
    msm_len = prepare_scalars_circuit.msm_len
    prepare_scalars_function_name = (
        f"{curve_id.name}_{prepare_scalars_circuit.name.upper()}"
    )
    prepare_scalars_code, prepare_scalars_function_name = (
        prepare_scalars_circuit.circuit.compile_circuit(
            function_name=prepare_scalars_function_name, pub=True, generic_modulus=True
        )
    )

    # Generate consistency circuits for ZK
    consistency_circuits = []
    if is_zk:
        for circuit_class in [
            ZKEvalsConsistencyInitCircuit,
            ZKEvalsConsistencyLoopCircuit,
            ZKEvalsConsistencyDoneCircuit,
        ]:
            circuit = circuit_class(vk)
            function_name = f"{curve_id.name}_{circuit.name.upper()}"
            circuit_code, function_name = circuit.circuit.compile_circuit(
                function_name=function_name, pub=True, generic_modulus=True
            )
            consistency_circuits.append((circuit_code, function_name))

    # Combine all circuit code
    code += sumcheck_code + prepare_scalars_code
    if is_zk:
        code += "".join(circuit_code for circuit_code, _ in consistency_circuits)
    code += get_circuit_definition_impl_template(len(scalar_indexes))

    is_on_curve_code = """
#[inline(never)]
pub fn is_on_curve_bn254(p: G1Point, modulus: CircuitModulus) -> bool {
    // INPUT stack
    // y^2 = x^3 + 3
    let (in0, in1) = (CE::<CI<0>> {}, CE::<CI<1>> {});
    let y2 = circuit_mul(in1, in1);
    let x2 = circuit_mul(in0, in0);
    let x3 = circuit_mul(in0, x2);
    let y2_minus_x3 = circuit_sub(y2, x3);

    let mut circuit_inputs = (y2_minus_x3,).new_inputs();
    // Prefill constants:

    // Fill inputs:
    circuit_inputs = circuit_inputs.next_2(p.x); // in0
    circuit_inputs = circuit_inputs.next_2(p.y); // in1

    let outputs = circuit_inputs.done_2().eval(modulus).unwrap();
    let zero_check: u384 = outputs.get_output(y2_minus_x3);
    return zero_check == u384{limb0: 3, limb1: 0, limb2: 0, limb3: 0};
}
    """
    code += is_on_curve_code
    is_on_curve_function_name = f"is_on_curve_bn254"

    return (
        code,
        sumcheck_function_name,
        prepare_scalars_function_name,
        [func_name for _, func_name in consistency_circuits] if is_zk else [],
        scalar_indexes,
        msm_len,
        is_on_curve_function_name,
    )


def _gen_contract_header(flavor: str, is_zk: bool, function_names: list[str]) -> str:
    """Generate the contract header and imports based on flavor and ZK status."""
    contract_name = f"Ultra{flavor}{'ZK' if is_zk else ''}HonkVerifier"
    trait_name = f"I{contract_name}"
    endpoint_name = f"verify_ultra_{flavor.lower()}{'_zk' if is_zk else ''}_honk_proof"
    imports_str = ", ".join(function_names)
    proof_struct_name = f"{('ZK' if is_zk else '') + 'HonkProof'}"
    header = f"""
use super::honk_verifier_constants::{{vk, VK_HASH, precomputed_lines}};
use super::honk_verifier_circuits::{{{imports_str}}};

#[starknet::interface]
pub trait {trait_name}<TContractState> {{
    fn {endpoint_name}(
        self: @TContractState,
        full_proof_with_hints: Span<felt252>,
    ) -> Option<Span<u256>>;
}}

#[starknet::contract]
mod {contract_name} {{
    use garaga::definitions::{{G1Point, G1G2Pair, BN254_G1_GENERATOR, get_BN254_modulus, get_GRUMPKIN_modulus, u288, u384, get_eigenvalue, get_third_root_of_unity, get_min_one_order, get_nG_glv_fake_glv}};
    use garaga::pairing_check::{{multi_pairing_check_bn254_2P_2F, MPCheckHintBN254}};
    use garaga::ec_ops::{{G1PointTrait, _ec_safe_add, _scalar_mul_glv_and_fake_glv, GlvFakeGlvHint}};
    use garaga::circuits::ec;
    use super::{{vk, VK_HASH, precomputed_lines, {imports_str}}};
    use garaga::utils::noir::{{{proof_struct_name}, G2_POINT_KZG_1, G2_POINT_KZG_2}};
    use garaga::utils::noir::honk_transcript::{{Point256IntoCircuitPoint, {flavor}HasherState}};
    use garaga::utils::noir::{'zk_' if is_zk else ''}honk_transcript::{{{('ZK' if is_zk else '') + 'HonkTranscriptTrait'}, {'ZK_' if is_zk else ''}BATCHED_RELATION_PARTIAL_LENGTH}};
    use garaga::core::circuit::{{U32IntoU384, u288IntoCircuitInputValue, U64IntoU384, {'u256_to_u384, ' if is_zk else ''}}};
    use core::num::traits::Zero;
    use core::poseidon::hades_permutation;

    #[storage]
    struct Storage {{}}

    #[derive(Drop, Serde)]
    struct FullProof {{
        proof: {proof_struct_name},
        msm_hint: Span<felt252>,
        kzg_hint:MPCheckHintBN254,
    }}

    #[abi(embed_v0)]
    impl {trait_name} of super::{trait_name}<ContractState> {{
        fn {endpoint_name}(
            self: @ContractState,
            full_proof_with_hints: Span<felt252>,
        ) -> Option<Span<u256>> {{
            // DO NOT EDIT THIS FUNCTION UNLESS YOU KNOW WHAT YOU ARE DOING.
            // This function returns an Option for the public inputs if the proof is valid.
            // If the proof is invalid, the execution will either fail or return None.
            // Read the documentation to learn how to generate the full_proof_with_hints array given a proof and a verifying key.
            let mut full_proof_with_hints = full_proof_with_hints;
            let full_proof = Serde::<FullProof>::deserialize(ref full_proof_with_hints).expect('deserialization failed');
            let mod_bn = get_BN254_modulus();
            let mod_grumpkin = get_GRUMPKIN_modulus();

"""
    return header


def _gen_constants_code(vk: HonkVk) -> str:
    """Generate the constants code for the contract."""
    lines = precompute_lines([G2_POINT_KZG_1, G2_POINT_KZG_2])
    precomputed_lines = StructArray(
        name="lines",
        elmts=[
            G2Line(name=f"line{i}", elmts=lines[i : i + 4])
            for i in range(0, len(lines), 4)
        ],
    )
    return f"""
use garaga::definitions::{{G1Point, G2Line, u384, u288}};
use garaga::utils::noir::HonkVk;

// _vk_hash = keccak256(vk_bytes)
// vk_hash = hades_permutation(_vk_hash.low, _vk_hash.high, 2)
pub const VK_HASH: felt252 = {hex(vk.vk_hash)};
{vk.serialize_to_cairo()}\n
pub const precomputed_lines: [G2Line; {len(precomputed_lines)//4}] = {precomputed_lines.serialize(raw=True, const=True)};
"""


def _get_msm_points_array_code(zk: bool, log_n: int) -> tuple[str, (int, int)]:
    # Common points that are shared between ZK and non-ZK versions
    vk_points = [
        "vk.qm",
        "vk.qc",
        "vk.ql",
        "vk.qr",
        "vk.qo",
        "vk.q4",
        "vk.qLookup",
        "vk.qArith",
        "vk.qDeltaRange",
        "vk.qElliptic",
        "vk.qAux",
        "vk.qPoseidon2External",
        "vk.qPoseidon2Internal",
        "vk.s1",
        "vk.s2",
        "vk.s3",
        "vk.s4",
        "vk.id1",
        "vk.id2",
        "vk.id3",
        "vk.id4",
        "vk.t1",
        "vk.t2",
        "vk.t3",
        "vk.t4",
        "vk.lagrange_first",
        "vk.lagrange_last",
    ]

    next_proof_point_counter = 1
    proof_points = (
        [
            f"full_proof.proof.gemini_masking_poly.into(), // Proof point {next_proof_point_counter}"
        ]
        if zk
        else []
    )
    next_proof_point_counter += 1 if zk else 0

    proof_points += [
        f"full_proof.proof.w1.into(), // Proof point {next_proof_point_counter}",
        f"full_proof.proof.w2.into(), // Proof point {next_proof_point_counter+1}",
        f"full_proof.proof.w3.into(), // Proof point {next_proof_point_counter+2}",
        f"full_proof.proof.w4.into(), // Proof point {next_proof_point_counter+3}",
        f"full_proof.proof.z_perm.into(), // Proof point {next_proof_point_counter+4}",
        f"full_proof.proof.lookup_inverses.into(), // Proof point {next_proof_point_counter+5}",
        f"full_proof.proof.lookup_read_counts.into(), // Proof point {next_proof_point_counter+6}",
        f"full_proof.proof.lookup_read_tags.into(), // Proof point {next_proof_point_counter+7}",
    ]
    next_proof_point_counter += 8

    # Combine all points
    all_points = vk_points + proof_points

    points_str = ",\n".join(all_points)
    code = "// Starts with 1 * shplonk_q, not included in msm\n"
    code += f"""            let mut _points: Array<G1Point> = array![{points_str}\n];

            for gem_comm in full_proof.proof.gemini_fold_comms {{
                _points.append((*gem_comm).into());
            }}; // log_n -1 = {log_n-1} points || Proof points {next_proof_point_counter}-{next_proof_point_counter+(log_n-1) - 1}"""
    next_proof_point_counter += log_n - 1
    # Add ZK-specific commitments
    if zk:
        code += f"""
            for lib_comm in full_proof.proof.libra_commitments {{
                _points.append((*lib_comm).into());
            }};// 3 points || Proof points {next_proof_point_counter}-{next_proof_point_counter+3-1}"""
        next_proof_point_counter += 3
    # Add common final points
    code += f"""
            _points.append(full_proof.proof.kzg_quotient.into()); // Proof point {next_proof_point_counter}
            _points.append(BN254_G1_GENERATOR);

            let mut points = _points.span();"""

    proof_point_counter = next_proof_point_counter

    n_vk_points = len(vk_points)

    return code, (n_vk_points, proof_point_counter)


def _gen_honk_verifier_files(
    vk: str | Path | HonkVk | bytes,
    system: ProofSystem = ProofSystem.UltraKeccakHonk,
) -> tuple[str, str, str]:
    vk, flavor = setup_vk_flavor(system, vk)

    curve_id = CurveID.GRUMPKIN

    constants_code = _gen_constants_code(vk)

    (
        circuits_code,
        sumcheck_function_name,
        prepare_scalars_function_name,
        consistency_function_names,
        scalar_indexes,
        msm_len,
        is_on_curve_function_name,
    ) = _gen_circuits_code(vk, False)

    scalars_tuple = ",\n            ".join(f"scalar_{idx}" for idx in scalar_indexes)
    scalars_tuple_into = [f"scalar_{idx}" for idx in scalar_indexes]

    scalars_tuple_into.append("transcript.shplonk_z.into()")
    # Swap position of last two elements of scalars_tuple_into :
    scalars_tuple_into = scalars_tuple_into[:-2] + scalars_tuple_into[-2:][::-1]
    scalars_tuple_into = ",\n            ".join(scalars_tuple_into)

    # Generate contract header
    contract_header = _gen_contract_header(
        flavor=flavor,
        is_zk=False,
        function_names=[
            sumcheck_function_name,
            prepare_scalars_function_name,
            is_on_curve_function_name,
        ],
    )

    points_code, (n_vk_points, n_proof_points) = _get_msm_points_array_code(
        zk=False, log_n=vk.log_circuit_size
    )

    contract_code = f"""{contract_header}
            let (transcript, transcript_state, base_rlc) = HonkTranscriptTrait::from_proof::<{flavor}HasherState>(vk.circuit_size, vk.public_inputs_size, vk.public_inputs_offset, full_proof.proof);
            let log_n = vk.log_circuit_size;
            let (sum_check_rlc, honk_check) = {sumcheck_function_name}(
                p_public_inputs: full_proof.proof.public_inputs,
                p_pairing_point_object: full_proof.proof.pairing_point_object,
                p_public_inputs_offset: vk.public_inputs_offset.into(),
                sumcheck_univariates_flat: full_proof.proof.sumcheck_univariates.slice(0, log_n * BATCHED_RELATION_PARTIAL_LENGTH),
                sumcheck_evaluations: full_proof.proof.sumcheck_evaluations,
                tp_sum_check_u_challenges: transcript.sum_check_u_challenges.span().slice(0, log_n),
                tp_gate_challenges: transcript.gate_challenges.span().slice(0, log_n),
                tp_eta_1: transcript.eta,
                tp_eta_2: transcript.eta_two,
                tp_eta_3: transcript.eta_three,
                tp_beta: transcript.beta,
                tp_gamma: transcript.gamma,
                tp_base_rlc: base_rlc.into(),
                tp_alphas: transcript.alphas.span(),
                modulus: mod_grumpkin,
            );

        let (
            {scalars_tuple},
        ) =
            {prepare_scalars_function_name}(
            p_sumcheck_evaluations: full_proof.proof.sumcheck_evaluations,
            p_gemini_a_evaluations: full_proof.proof.gemini_a_evaluations,
            tp_gemini_r: transcript.gemini_r.into(),
            tp_rho: transcript.rho.into(),
            tp_shplonk_z: transcript.shplonk_z.into(),
            tp_shplonk_nu: transcript.shplonk_nu.into(),
            tp_sum_check_u_challenges: transcript.sum_check_u_challenges.span().slice(0, log_n),
            modulus: mod_grumpkin,
        );

            {points_code}

            let mut scalars: Span<u384> = array![{scalars_tuple_into}].span();

            {get_msm_kzg_template(msm_len, n_vk_points, n_proof_points, is_on_curve_function_name)}

            if sum_check_rlc.is_zero() && honk_check.is_zero() && kzg_check {{
                return Option::Some(full_proof.proof.public_inputs);
            }} else {{
                return Option::None;
            }}
        }}
    }}
}}


    """

    return constants_code, circuits_code, contract_code


def _gen_zk_honk_verifier_files(
    vk: str | Path | HonkVk | bytes,
    system: ProofSystem = ProofSystem.UltraKeccakZKHonk,
) -> str:
    vk, flavor = setup_vk_flavor(system, vk)

    curve_id = CurveID.GRUMPKIN

    constants_code = _gen_constants_code(vk)

    (
        circuits_code,
        sumcheck_function_name,
        prepare_scalars_function_name,
        consistency_function_names,
        scalar_indexes,
        msm_len,
        is_on_curve_function_name,
    ) = _gen_circuits_code(vk, True)

    points_code, (n_vk_points, n_proof_points) = _get_msm_points_array_code(
        zk=True, log_n=vk.log_circuit_size
    )

    scalars_tuple = ",\n            ".join(f"scalar_{idx}" for idx in scalar_indexes)
    scalars_tuple_into = [f"scalar_{idx}" for idx in scalar_indexes]

    scalars_tuple_into.append("transcript.shplonk_z.into()")
    # Swap position of last two elements of scalars_tuple_into (KZG quotient & BN254 G1 generator) :
    scalars_tuple_into = scalars_tuple_into[:-2] + scalars_tuple_into[-2:][::-1]
    # Move first scalar after the first vk points [a, b, c, d, e] -> [b,c,d,a,e] if [b,c,d] are vk points:
    scalars_tuple_into = (
        scalars_tuple_into[1 : n_vk_points + 1]
        + scalars_tuple_into[:1]
        + scalars_tuple_into[n_vk_points + 1 :]
    )

    scalars_tuple_into = ",\n            ".join(scalars_tuple_into)

    # Generate contract header
    contract_header = _gen_contract_header(
        flavor=flavor,
        is_zk=True,
        function_names=[
            sumcheck_function_name,
            prepare_scalars_function_name,
            consistency_function_names[0],
            consistency_function_names[1],
            consistency_function_names[2],
            is_on_curve_function_name,
        ],
    )

    contract_code = f"""{contract_header}
            let (transcript, transcript_state, base_rlc) = ZKHonkTranscriptTrait::from_proof::<{flavor}HasherState>(vk.circuit_size, vk.public_inputs_size, vk.public_inputs_offset, full_proof.proof);
            let log_n = vk.log_circuit_size;
            let (sum_check_rlc, honk_check) = {sumcheck_function_name}(
                p_public_inputs: full_proof.proof.public_inputs,
                p_pairing_point_object: full_proof.proof.pairing_point_object,
                p_public_inputs_offset: vk.public_inputs_offset.into(),
                libra_sum: u256_to_u384(full_proof.proof.libra_sum),
                sumcheck_univariates_flat: full_proof.proof.sumcheck_univariates.slice(0, log_n * ZK_BATCHED_RELATION_PARTIAL_LENGTH),
                sumcheck_evaluations: full_proof.proof.sumcheck_evaluations,
                libra_evaluation: u256_to_u384(full_proof.proof.libra_evaluation),
                tp_sum_check_u_challenges: transcript.sum_check_u_challenges.span().slice(0, log_n),
                tp_gate_challenges: transcript.gate_challenges.span().slice(0, log_n),
                tp_eta_1: transcript.eta.into(),
                tp_eta_2: transcript.eta_two.into(),
                tp_eta_3: transcript.eta_three.into(),
                tp_beta: transcript.beta.into(),
                tp_gamma: transcript.gamma.into(),
                tp_base_rlc: base_rlc.into(),
                tp_alphas: transcript.alphas.span(),
                tp_libra_challenge: transcript.libra_challenge.into(),
                modulus: mod_grumpkin,
            );

            const CONST_PROOF_SIZE_LOG_N: usize = {CONST_PROOF_SIZE_LOG_N};
            let (mut challenge_poly_eval, mut root_power_times_tp_gemini_r) = {consistency_function_names[0]}(
                tp_gemini_r: transcript.gemini_r.into(),
                modulus: mod_grumpkin,
            );
            for i in 0..CONST_PROOF_SIZE_LOG_N {{
                let (new_challenge_poly_eval, new_root_power_times_tp_gemini_r) = {consistency_function_names[1]}(
                    challenge_poly_eval: challenge_poly_eval,
                    root_power_times_tp_gemini_r: root_power_times_tp_gemini_r,
                    tp_sumcheck_u_challenge: (*transcript.sum_check_u_challenges.at(i)).into(),
                    modulus: mod_grumpkin,
                );
                challenge_poly_eval = new_challenge_poly_eval;
                root_power_times_tp_gemini_r = new_root_power_times_tp_gemini_r;
            }};
            let (vanishing_check, diff_check) = {consistency_function_names[2]}(
                p_libra_evaluation: u256_to_u384(full_proof.proof.libra_evaluation),
                p_libra_poly_evals: full_proof.proof.libra_poly_evals,
                tp_gemini_r: transcript.gemini_r.into(),
                challenge_poly_eval: challenge_poly_eval,
                root_power_times_tp_gemini_r: root_power_times_tp_gemini_r,
                modulus: mod_grumpkin,
            );

        let (
            {scalars_tuple},
        ) =
            {prepare_scalars_function_name}(
            p_sumcheck_evaluations: full_proof.proof.sumcheck_evaluations,
            p_gemini_masking_eval: u256_to_u384(full_proof.proof.gemini_masking_eval),
            p_gemini_a_evaluations: full_proof.proof.gemini_a_evaluations,
            p_libra_poly_evals: full_proof.proof.libra_poly_evals,
            tp_gemini_r: transcript.gemini_r.into(),
            tp_rho: transcript.rho.into(),
            tp_shplonk_z: transcript.shplonk_z.into(),
            tp_shplonk_nu: transcript.shplonk_nu.into(),
            tp_sum_check_u_challenges: transcript.sum_check_u_challenges.span().slice(0, log_n),
            modulus: mod_grumpkin,
        );

            {points_code}
            let mut scalars: Span<u384> = array![{scalars_tuple_into}].span();

            {get_msm_kzg_template(msm_len, n_vk_points, n_proof_points, is_on_curve_function_name)}
            if sum_check_rlc.is_zero() && honk_check.is_zero() && !vanishing_check.is_zero() && diff_check.is_zero() && kzg_check {{
                return Option::Some(full_proof.proof.public_inputs);
            }} else {{
                return Option::None;
            }}
        }}
    }}
}}


    """
    return constants_code, circuits_code, contract_code


def _write_and_format_project_files(
    output_folder_path: str,
    output_folder_name: str,
    cli_mode: bool,
    constants_code: str,
    circuits_code: str,
    contract_code: str,
):
    create_directory(output_folder_path)
    src_dir = os.path.join(output_folder_path, "src")
    create_directory(src_dir)

    with open(os.path.join(output_folder_path, ".tool-versions"), "w") as f:
        f.write(f"scarb {CAIRO_VERSION}\n")

    with open(os.path.join(src_dir, "honk_verifier_constants.cairo"), "w") as f:
        f.write(constants_code)

    with open(os.path.join(src_dir, "honk_verifier_circuits.cairo"), "w") as f:
        f.write(circuits_code)

    with open(os.path.join(src_dir, "honk_verifier.cairo"), "w") as f:
        f.write(contract_code)

    with open(os.path.join(output_folder_path, "Scarb.toml"), "w") as f:
        f.write(get_scarb_toml_file(output_folder_name, cli_mode))

    with open(os.path.join(src_dir, "lib.cairo"), "w") as f:
        f.write(
            """
mod honk_verifier;
mod honk_verifier_constants;
mod honk_verifier_circuits;
"""
        )
    subprocess.run(["scarb", "fmt", f"."], check=True, cwd=output_folder_path)


if __name__ == "__main__":
    import concurrent.futures

    parser = argparse.ArgumentParser(description="Generate HONK verifier contracts")
    parser.add_argument(
        "--max-log-n",
        action="store_true",
        help="Generate contracts with maximum log N size",
    )
    args = parser.parse_args()

    VK_PATH = (
        "hydra/garaga/starknet/honk_contract_generator/examples/vk_ultra_keccak.bin"
    )
    CONTRACTS_FOLDER = "src/contracts/autogenerated/"  # Do not change this

    if not args.max_log_n:
        systems = [
            ProofSystem.UltraKeccakHonk,
            ProofSystem.UltraStarknetHonk,
            ProofSystem.UltraKeccakZKHonk,
            ProofSystem.UltraStarknetZKHonk,
        ]

        with concurrent.futures.ProcessPoolExecutor() as executor:
            futures = []
            for system in systems:
                match system:
                    case ProofSystem.UltraKeccakHonk:
                        flavor = "keccak"
                    case ProofSystem.UltraStarknetHonk:
                        flavor = "starknet"
                    case ProofSystem.UltraKeccakZKHonk:
                        flavor = "keccak_zk"
                    case ProofSystem.UltraStarknetZKHonk:
                        flavor = "starknet_zk"
                    case _:
                        raise ValueError(f"Proof system {system} not compatible")

                FOLDER_NAME = f"noir_ultra_{flavor}_honk_example"  # '_curve_id' is appended in the end.

                # Submit the task to the executor
                futures.append(
                    executor.submit(
                        gen_honk_verifier,
                        VK_PATH,
                        CONTRACTS_FOLDER,
                        FOLDER_NAME,
                        system,
                    )
                )

            # Wait for all futures to complete
            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()  # This will raise an exception if the function raised
                except Exception as e:
                    print(f"An error occurred: {e}")

    if args.max_log_n:
        # NOTE : each additional public input increase bytecode by ~ 21 felts.
        #       each additional circuit size increase bytecode by ~ 372 felts.
        vk = HonkVk.from_bytes(open(VK_PATH, "rb").read())
        vk.log_circuit_size = MAX_LOG_N
        vk.public_inputs_size = 4 + PAIRING_POINT_OBJECT_LENGTH

        gen_honk_verifier(
            vk,
            CONTRACTS_FOLDER,
            "ultra_keccak_zk_honk_max_log_n",
            system=ProofSystem.UltraKeccakZKHonk,
        )



"""
Auto-generated constants file. Do not edit manually.
Generated from constants.json by tools/make/generate_constants.py
"""

# RISC0 Constants
# https://github.com/risc0/risc0-ethereum/blob/main/contracts/src/groth16/ControlID.sol
# release 2.0
RISC0_CONTROL_ROOT = 0x539032186827B06719244873B17B2D4C122E2D02CFB1994FE958B2523B844576
RISC0_BN254_CONTROL_ID = (
    0x04446E66D300EB7FB45C9726BB53C793DDA407A62E9601618BB43C5C14657AC0
)

# SP1 Constants
# https://github.com/succinctlabs/sp1-contracts/blob/main/contracts/src/v4.0.0-rc.3/SP1VerifierGroth16.sol
SP1_VERIFIER_VERSION: str = "v4.0.0-rc.3"
SP1_VERIFIER_HASH: bytes = bytes.fromhex(
    "11b6a09d63d255ad425ee3a7f6211d5ec63fbde9805b40551c3136275b6f4eb4"
)

# Additional RISC0 constants for internal use
RISC0_SYSTEM_STATE_ZERO_DIGEST = (
    "0xA3ACC27117418996340B84E5A90F3EF4C49D22C79E44AAD822EC9C313E1EB8E2"
)
RISC0_TAG_DIGEST = "0xcb1fefcd1f2d9a64975cbbbf6e161e2914434b0cbb9960b84df5d717e86b48af"
RISC0_OUTPUT_TAG = "0x77eafeb366a78b47747de0d7bb176284085ff5564887009a5be63da32d3559d4"



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Type, Union

import garaga.modulo_circuit_structs as structs
from garaga.definitions import BLS12_381_ID
from garaga.modulo_circuit import ModuloCircuit, PyFelt
from garaga.modulo_circuit_structs import (
    Cairo1SerializableStruct,
    G2Line,
    G2PointCircuit,
    u384,
    u384Array,
)
from garaga.precompiled_circuits.compilable_circuits.base import (
    BaseEXTFCircuit,
    BaseModuloCircuit,
)
from garaga.precompiled_circuits.miller_tower import MillerTowerCircuit


def split_4_sized_object_into_tuple_of_2_size(
    input: Union[List[PyFelt], Tuple[PyFelt, PyFelt, PyFelt, PyFelt]]
) -> Optional[Tuple[List[PyFelt], List[PyFelt]]]:
    if input is None:
        return None
    assert len(input) == 4, f"Expected input of length 4, got {len(input)}"
    return (list(input[0:2]), list(input[2:4]))


def parse_precomputed_g1_consts_and_g2_points(
    circuit: MillerTowerCircuit,
    vars: dict,
    n_pairs: int,
    bit_1: bool = False,
) -> list[tuple[list[PyFelt], list[PyFelt]]]:
    current_points = []
    q_or_q_neg_points = [] if bit_1 else None
    for i in range(n_pairs):
        circuit.yInv.append(vars[f"yInv_{i}"])
        circuit.xNegOverY.append(vars[f"xNegOverY_{i}"])
        current_points.append(
            split_4_sized_object_into_tuple_of_2_size(vars.get(f"Q_{i}", None))
        )  # Return empty list if not present
        if bit_1:
            q_or_q_neg_points.append(
                split_4_sized_object_into_tuple_of_2_size(
                    vars.get(f"Q_or_Q_neg_{i}", None)
                )  # Return empty list if not present
            )

    return current_points, q_or_q_neg_points


class BaseTowerMillerLoop(BaseEXTFCircuit, ABC):
    def __init__(
        self,
        name,
        curve_id,
        auto_run=True,
        compilation_mode=1,
        n_pairs: int = 1,
    ):
        assert compilation_mode == 1, "Compilation mode 1 is required for this circuit"
        self.n_pairs = n_pairs
        super().__init__(
            name=name,
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )
        self.generic_over_curve = True

    def _initialize_circuit(self):
        return MillerTowerCircuit(
            self.name,
            self.curve_id,
            n_pairs=self.n_pairs,
            compilation_mode=self.compilation_mode,
        )

    @property
    @abstractmethod
    def input_map(
        self,
    ) -> Dict[
        str,
        Union[
            Type[Cairo1SerializableStruct], Tuple[Type[Cairo1SerializableStruct], int]
        ],
    ]:
        """
        Define the input map for the circuit in a dict.
        The key will be the name of the input variable, also used in the signature of the compiled Cairo code.
        The value will be either a Cairo1SerializableStruct type (which defines the struct in the Cairo code),
            or a tuple of the type and its size (for Array-like types).
        The reason behind this is that each Cairo1SerializableStruct defines the __len__ method, but for the
        array-like structs we need to specify the size in advance.
        """

    def _base_input_map(self, bit_type: str) -> dict:
        """
        Base input map for the bit 0 and 1
        """
        input_map = {}

        for k in range(self.n_pairs):
            input_map[f"yInv_{k}"] = u384
            input_map[f"xNegOverY_{k}"] = u384
            input_map[f"Q_{k}"] = G2PointCircuit
            if bit_type == "1":
                input_map[f"Q_or_Q_neg_{k}"] = G2PointCircuit

        # Add common inputs
        input_map["M_i"] = structs.E12T

        return input_map

    def _process_input(self, circuit: MillerTowerCircuit, input: list[PyFelt]) -> dict:
        """
        Method responsible for deserializing the input list of elements into the variables in the input map,
        and writing them to the circuit.
        The input list is expected to be in the same order as the input map.
        Since we use Python 3.10, the input map dict is ordered.
        Returns a vars dict with the same keys as the input map, but with the values being the instances of the structs,
        each struct holding ModuloCircuitElement(s).
        """
        vars = {}
        for name, struct_info in self.input_map.items():
            if isinstance(struct_info, tuple) and struct_info[0] == u384Array:
                # Handle u384Array with specified size
                struct_type, size = struct_info
                vars[name] = circuit.write_struct(
                    struct_type(name, elmts=[input.pop(0) for _ in range(size)])
                )
            elif struct_info == G2Line:
                # Internally appended to the precomputed_lines flattened list
                circuit.precomputed_lines.extend(
                    circuit.write_struct(
                        G2Line(name, elmts=[input.pop(0) for _ in range(4)])
                    )
                )
            else:
                struct_type = struct_info
                # For other types, create a temporary instance to get its length
                temp_instance = struct_type(
                    name, None
                )  # Temporary instance with a dummy element
                size = len(temp_instance)
                if size == 1:
                    vars[name] = circuit.write_struct(
                        struct_type(name, elmts=[input.pop(0)])
                    )
                else:
                    # Convert to tuple so that it is Hashable
                    vars[name] = tuple(
                        circuit.write_struct(
                            struct_type(name, elmts=[input.pop(0) for _ in range(size)])
                        )
                    )
        assert len(input) == 0, f"Expected input of length 0, got {len(input)}"
        # Create the precomputed lines generator
        circuit._precomputed_lines_generator = (
            circuit._create_precomputed_lines_generator()
        )
        return vars

    def build_input(self) -> list[PyFelt]:
        """
        Extends the base method of BaseModuloCircuit, by reading the input map and returning a list of random elements of the total expected size.
        """
        total_elements = 0
        for name, struct_info in self.input_map.items():
            if isinstance(struct_info, tuple):
                # Array-like case
                _, size = struct_info
                total_elements += size
            else:
                # For other structs, create a temporary instance to get its length
                temp_instance = struct_info(name, None)
                total_elements += len(temp_instance)

        return [self.field.random() for _ in range(total_elements)]

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = self._initialize_circuit()

        vars = self._process_input(circuit, input)
        return self._execute_circuit_logic(circuit, vars)

    @abstractmethod
    def _execute_circuit_logic(self, circuit, vars) -> ModuloCircuit:
        """
        Implement the circuit logic using the processed input variables.
        """

    def _execute_circuit_bit_logic_base(
        self, circuit: MillerTowerCircuit, vars, bit_type
    ):
        n_pairs = self.n_pairs
        assert n_pairs >= 1, f"n_pairs must be >= 1, got {n_pairs}"

        current_points, q_or_q_neg_points = parse_precomputed_g1_consts_and_g2_points(
            circuit, vars, n_pairs, bit_1=(bit_type == "1")
        )

        Mi_sq = circuit.fp12_square(vars["M_i"])

        new_points, Mi_plus_one = self._process_points(
            circuit, current_points, q_or_q_neg_points, Mi_sq, bit_type
        )

        self._extend_output(circuit, new_points, Mi_plus_one)

        return circuit

    def _process_points(
        self,
        circuit: MillerTowerCircuit,
        current_points,
        q_or_q_neg_points,
        sum_i_prod_k_P,
        bit_type,
    ):
        new_points = []
        if bit_type == "0":
            for k in range(self.n_pairs):
                T, (lineR0, lineR1) = circuit._double(current_points[k], k)
                line = circuit.eval_tower_line(
                    lineR0 + lineR1, circuit.yInv[k], circuit.xNegOverY[k]
                )
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [line], k
                )
                new_points.append(T)
        elif bit_type == "1":
            for k in range(self.n_pairs):
                # T, l1, l2 = circuit.double_and_add_step(
                #     current_points[k], q_or_q_neg_points[k], k
                # )
                T, (line1R0, line1R1), (line2R0, line2R1) = circuit._double_and_add(
                    current_points[k], q_or_q_neg_points[k], k
                )
                line1 = circuit.eval_tower_line(
                    line1R0 + line1R1, circuit.yInv[k], circuit.xNegOverY[k]
                )
                line2 = circuit.eval_tower_line(
                    line2R0 + line2R1, circuit.yInv[k], circuit.xNegOverY[k]
                )
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [line1, line2], k
                )
                new_points.append(T)

        return new_points, sum_i_prod_k_P

    def _multiply_line_evaluations(
        self,
        circuit: MillerTowerCircuit,
        sum_i_prod_k_P,
        lines,
        k,
    ):
        for i, l in enumerate(lines):
            sum_i_prod_k_P = circuit.mul_by_line_tower(sum_i_prod_k_P, l)
        return sum_i_prod_k_P

    def _extend_output(self, circuit: ModuloCircuit, new_points, Mi_plus_one):
        circuit.exact_output_refs_needed = []
        for i, point in enumerate(new_points):
            sum_coords = circuit.sum(
                [point[0][0], point[0][1], point[1][0], point[1][1]]
            )
            circuit.exact_output_refs_needed.append(sum_coords)

            circuit.extend_struct_output(
                G2PointCircuit(
                    name=f"Q{i}",
                    elmts=[
                        point[0][0],
                        point[0][1],
                        point[1][0],
                        point[1][1],
                    ],
                )
            )

        circuit.extend_struct_output(
            structs.E12T(name="Mi_plus_one", elmts=Mi_plus_one)
        )
        circuit.exact_output_refs_needed.extend(Mi_plus_one)


class TowerMillerBit0(BaseTowerMillerLoop):
    def __init__(
        self,
        curve_id: int,
        n_pairs: int,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ):
        super().__init__(
            name=f"tower_miller_bit0_{n_pairs}P",
            curve_id=curve_id,
            n_pairs=n_pairs,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    @property
    def input_map(self):
        return self._base_input_map("0")

    def _execute_circuit_logic(self, circuit, vars) -> ModuloCircuit:
        return self._execute_circuit_bit_logic_base(circuit, vars, "0")


class TowerMillerBit1(BaseTowerMillerLoop):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        n_pairs: int = 1,
        compilation_mode: int = 1,
    ):
        assert compilation_mode == 1, "Compilation mode 1 is required for this circuit"
        super().__init__(
            name=f"tower_miller_bit1_{n_pairs}P",
            curve_id=curve_id,
            n_pairs=n_pairs,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    @property
    def input_map(self):
        return self._base_input_map("1")

    def _execute_circuit_logic(self, circuit, vars) -> ModuloCircuit:
        return self._execute_circuit_bit_logic_base(circuit, vars, "1")


class TowerMillerInitBit(BaseTowerMillerLoop):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        compilation_mode: int = 1,
        n_pairs: int = 3,
    ):
        super().__init__(
            name=f"tower_miller_init_bit_{n_pairs}P",
            curve_id=curve_id,
            n_pairs=n_pairs,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    @property
    def input_map(self):
        if self.curve_id == BLS12_381_ID:
            input_map = {}
            for k in range(self.n_pairs):
                input_map[f"yInv_{k}"] = u384
                input_map[f"xNegOverY_{k}"] = u384
                input_map[f"Q_{k}"] = G2PointCircuit
            return input_map
        else:
            raise NotImplementedError(f"Curve {self.curve_id} not implemented")

    def _execute_circuit_logic(
        self, circuit: MillerTowerCircuit, vars
    ) -> ModuloCircuit:
        n_pairs = self.n_pairs
        assert n_pairs == 1, f"Only implemented for 1 pair for now, got {n_pairs}"

        current_points, q_or_q_neg_points = parse_precomputed_g1_consts_and_g2_points(
            circuit, vars, n_pairs, bit_1=False
        )

        new_points = []
        for k in range(self.n_pairs):
            T, (line1R0, line1R1), (line2R0, line2R1) = circuit._triple(
                current_points[k], k
            )

            line1 = circuit.eval_tower_line(
                line1R0 + line1R1, circuit.yInv[k], circuit.xNegOverY[k]
            )
            line2 = circuit.eval_tower_line(
                line2R0 + line2R1, circuit.yInv[k], circuit.xNegOverY[k]
            )
            # result.C0.B0 = prodLines[0]
            # 		result.C0.B1 = prodLines[1]
            # 		result.C0.B2 = prodLines[2]
            # 		result.C1.B1 = prodLines[3]
            # 		result.C1.B2 = prodLines[4]
            ll = circuit.mul_line_by_line_tower(line1, line2)

            new_points.append(T)

        for i, point in enumerate(new_points):
            circuit.extend_struct_output(
                G2PointCircuit(
                    name=f"Q{i}",
                    elmts=[
                        point[0][0],
                        point[0][1],
                        point[1][0],
                        point[1][1],
                    ],
                )
            )

        circuit.extend_struct_output(structs.u384(name="c0b0a0", elmts=[ll[0]]))
        circuit.extend_struct_output(structs.u384(name="c0b0a1", elmts=[ll[1]]))
        circuit.extend_struct_output(structs.u384(name="c0b1a0", elmts=[ll[2]]))
        circuit.extend_struct_output(structs.u384(name="c0b1a1", elmts=[ll[3]]))
        circuit.extend_struct_output(structs.u384(name="c0b2a0", elmts=[ll[4]]))
        circuit.extend_struct_output(structs.u384(name="c0b2a1", elmts=[ll[5]]))
        circuit.extend_struct_output(structs.u384(name="c1b1a0", elmts=[ll[6]]))
        circuit.extend_struct_output(structs.u384(name="c1b1a1", elmts=[ll[7]]))
        circuit.extend_struct_output(structs.u384(name="c1b2a0", elmts=[ll[8]]))
        circuit.extend_struct_output(structs.u384(name="c1b2a1", elmts=[ll[9]]))

        return circuit


class TowerMillerFinalizeBN(BaseTowerMillerLoop):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        n_pairs: int = 1,
        compilation_mode: int = 1,
    ):

        super().__init__(
            name=f"tower_miller_finalize_bn_{n_pairs}P",
            curve_id=curve_id,
            n_pairs=n_pairs,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    @property
    def input_map(self):
        if self.curve_id == BLS12_381_ID:
            return {}
        input_map = {}

        for i in range(self.n_pairs):
            input_map[f"original_Q{i}"] = G2PointCircuit
            input_map[f"yInv_{i}"] = u384
            input_map[f"xNegOverY_{i}"] = u384
            input_map[f"Q_{i}"] = G2PointCircuit

        input_map["Mi"] = structs.E12T

        return input_map

    def _execute_circuit_logic(self, circuit: MillerTowerCircuit, vars):
        if self.curve_id == BLS12_381_ID:
            return circuit

        n_pairs = self.n_pairs
        current_points = []
        for k in range(n_pairs):
            circuit.Q.append(
                split_4_sized_object_into_tuple_of_2_size(
                    vars.get(f"original_Q{k}", None)
                )
            )
            circuit.yInv.append(vars[f"yInv_{k}"])
            circuit.xNegOverY.append(vars[f"xNegOverY_{k}"])
            current_points.append(
                split_4_sized_object_into_tuple_of_2_size(vars.get(f"Q_{k}", None))
            )

        Mi = vars["Mi"]

        lines = circuit._bn254_finalize_step(current_points)

        for l in lines:
            ((l1r0, l1r1), (l2r0, l2r1)) = l

            l1_eval = circuit.eval_tower_line(
                l1r0 + l1r1, circuit.yInv[k], circuit.xNegOverY[k]
            )
            l2_eval = circuit.eval_tower_line(
                l2r0 + l2r1, circuit.yInv[k], circuit.xNegOverY[k]
            )

            ll = circuit.mul_line_by_line_tower(l1_eval, l2_eval)

            ll_full = ll + [
                circuit.set_or_get_constant(0),
                circuit.set_or_get_constant(0),
            ]

            Mi = circuit.fp12_mul(Mi, ll_full)

        circuit.extend_struct_output(structs.E12T("Mi", elmts=Mi))
        return circuit


class E12TMulCircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__("e12t_mul", curve_id, auto_run, init_hash, compilation_mode)

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(12)])  # X
        input.extend([self.field.random() for _ in range(12)])  # Y

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> MillerTowerCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        X = circuit.write_struct(
            structs.E12T("X", elmts=[input.pop(0) for _ in range(12)])
        )
        Y = circuit.write_struct(
            structs.E12T("Y", elmts=[input.pop(0) for _ in range(12)])
        )
        res = circuit.fp12_mul(X, Y)
        circuit.extend_struct_output(structs.E12T("res", elmts=res))

        return circuit


class FP6NegCircuit(BaseModuloCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ):
        super().__init__("fp6_neg", curve_id, auto_run, compilation_mode)

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(6)])

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = ModuloCircuit(
            self.name,
            self.curve_id,
            generic_circuit=True,
            compilation_mode=self.compilation_mode,
        )

        fp6 = [
            circuit.write_struct(structs.u384(name="b0a0", elmts=[input.pop(0)])),
            circuit.write_struct(structs.u384(name="b0a1", elmts=[input.pop(0)])),
            circuit.write_struct(structs.u384(name="b1a0", elmts=[input.pop(0)])),
            circuit.write_struct(structs.u384(name="b1a1", elmts=[input.pop(0)])),
            circuit.write_struct(structs.u384(name="b2a0", elmts=[input.pop(0)])),
            circuit.write_struct(structs.u384(name="b2a1", elmts=[input.pop(0)])),
        ]

        res = [circuit.neg(fp6[i]) for i in range(6)]

        circuit.extend_struct_output(structs.u384(name="r0", elmts=[res[0]]))
        circuit.extend_struct_output(structs.u384(name="r1", elmts=[res[1]]))
        circuit.extend_struct_output(structs.u384(name="r2", elmts=[res[2]]))
        circuit.extend_struct_output(structs.u384(name="r3", elmts=[res[3]]))
        circuit.extend_struct_output(structs.u384(name="r4", elmts=[res[4]]))
        circuit.extend_struct_output(structs.u384(name="r5", elmts=[res[5]]))

        return circuit


class E12TInverseCircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(
            "e12t_inverse", curve_id, auto_run, init_hash, compilation_mode
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(12)])  # M

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        M = circuit.write_struct(
            structs.E12T("M", elmts=[input.pop(0) for _ in range(12)])
        )
        res = circuit.fp12_inverse(M)
        circuit.extend_struct_output(structs.E12T("res", elmts=res))

        return circuit


class E12TFrobeniusCircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 1,
    ):
        super().__init__(
            "e12t_frobenius", curve_id, auto_run, init_hash, compilation_mode
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(12)])  # M

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        M = circuit.write_struct(
            structs.E12T("M", elmts=[input.pop(0) for _ in range(12)])
        )
        res = circuit.fp12_frob(M)
        circuit.extend_struct_output(structs.E12T("res", elmts=res))

        return circuit


class E12TFrobeniusSquareCircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(
            "e12t_frobenius_square", curve_id, auto_run, init_hash, compilation_mode
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(12)])  # M

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        M = circuit.write_struct(
            structs.E12T("M", elmts=[input.pop(0) for _ in range(12)])
        )
        res = circuit.fp12_frob_square(M)
        circuit.extend_struct_output(structs.E12T("res", elmts=res))

        return circuit


class E12TFrobeniusCubeCircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(
            "e12t_frobenius_cube", curve_id, auto_run, init_hash, compilation_mode
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(12)])  # M

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        M = circuit.write_struct(
            structs.E12T("M", elmts=[input.pop(0) for _ in range(12)])
        )
        res = circuit.fp12_frob_cube(M)
        circuit.extend_struct_output(structs.E12T("res", elmts=res))

        return circuit


class E12TCyclotomicSquareCircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(
            "e12t_cyclotomic_square", curve_id, auto_run, init_hash, compilation_mode
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(12)])  # M

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        M = circuit.write_struct(
            structs.E12T("M", elmts=[input.pop(0) for _ in range(12)])
        )
        res = circuit.fp12_cyclotomic_square(M)
        circuit.extend_struct_output(structs.E12T("res", elmts=res))

        return circuit


class E12TCyclotomicSquareCompressedCircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(
            "e12t_cyclo_square_compressed",
            curve_id,
            auto_run,
            init_hash,
            compilation_mode,
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(8)])

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        xc0b1a0 = circuit.write_struct(
            structs.u384(name="xc0b1a0", elmts=[input.pop(0)])
        )
        xc0b1a1 = circuit.write_struct(
            structs.u384(name="xc0b1a1", elmts=[input.pop(0)])
        )
        xc0b2a0 = circuit.write_struct(
            structs.u384(name="xc0b2a0", elmts=[input.pop(0)])
        )
        xc0b2a1 = circuit.write_struct(
            structs.u384(name="xc0b2a1", elmts=[input.pop(0)])
        )
        xc1b0a0 = circuit.write_struct(
            structs.u384(name="xc1b0a0", elmts=[input.pop(0)])
        )
        xc1b0a1 = circuit.write_struct(
            structs.u384(name="xc1b0a1", elmts=[input.pop(0)])
        )
        xc1b2a0 = circuit.write_struct(
            structs.u384(name="xc1b2a0", elmts=[input.pop(0)])
        )
        xc1b2a1 = circuit.write_struct(
            structs.u384(name="xc1b2a1", elmts=[input.pop(0)])
        )
        res = circuit.fp12_cyclotomic_square_compressed(
            [xc0b1a0, xc0b1a1],
            [xc0b2a0, xc0b2a1],
            [xc1b0a0, xc1b0a1],
            [xc1b2a0, xc1b2a1],
        )
        assert len(res) == 8
        circuit.extend_struct_output(structs.u384(name="xc0b1a0", elmts=[res[0]]))
        circuit.extend_struct_output(structs.u384(name="xc0b1a1", elmts=[res[1]]))
        circuit.extend_struct_output(structs.u384(name="xc0b2a0", elmts=[res[2]]))
        circuit.extend_struct_output(structs.u384(name="xc0b2a1", elmts=[res[3]]))
        circuit.extend_struct_output(structs.u384(name="xc1b0a0", elmts=[res[4]]))
        circuit.extend_struct_output(structs.u384(name="xc1b0a1", elmts=[res[5]]))
        circuit.extend_struct_output(structs.u384(name="xc1b2a0", elmts=[res[6]]))
        circuit.extend_struct_output(structs.u384(name="xc1b2a1", elmts=[res[7]]))

        return circuit


class E12TDecompressKarabinaPtIZCircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(
            "e12t_decomp_karabina_I_Z", curve_id, auto_run, init_hash, compilation_mode
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(4)])

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        xc0b1a0 = circuit.write_struct(
            structs.u384(name="xc0b1a0", elmts=[input.pop(0)])
        )
        xc0b1a1 = circuit.write_struct(
            structs.u384(name="xc0b1a1", elmts=[input.pop(0)])
        )
        xc1b2a0 = circuit.write_struct(
            structs.u384(name="xc1b2a0", elmts=[input.pop(0)])
        )
        xc1b2a1 = circuit.write_struct(
            structs.u384(name="xc1b2a1", elmts=[input.pop(0)])
        )
        res = circuit.fp12_decompress_karabina_pt_I_c1b2_Z(
            [xc0b1a0, xc0b1a1], [xc1b2a0, xc1b2a1]
        )
        circuit.extend_struct_output(structs.u384(name="res0", elmts=[res[0]]))
        circuit.extend_struct_output(structs.u384(name="res1", elmts=[res[1]]))

        return circuit


class E12TDecompressKarabinaPtINZCircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(
            "e12t_decomp_karabina_I_NZ", curve_id, auto_run, init_hash, compilation_mode
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(8)])

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        xc0b1a0 = circuit.write_struct(
            structs.u384(name="xc0b1a0", elmts=[input.pop(0)])
        )
        xc0b1a1 = circuit.write_struct(
            structs.u384(name="xc0b1a1", elmts=[input.pop(0)])
        )
        xc0b2a0 = circuit.write_struct(
            structs.u384(name="xc0b2a0", elmts=[input.pop(0)])
        )
        xc0b2a1 = circuit.write_struct(
            structs.u384(name="xc0b2a1", elmts=[input.pop(0)])
        )
        xc1b0a0 = circuit.write_struct(
            structs.u384(name="xc1b0a0", elmts=[input.pop(0)])
        )
        xc1b0a1 = circuit.write_struct(
            structs.u384(name="xc1b0a1", elmts=[input.pop(0)])
        )
        xc1b2a0 = circuit.write_struct(
            structs.u384(name="xc1b2a0", elmts=[input.pop(0)])
        )
        xc1b2a1 = circuit.write_struct(
            structs.u384(name="xc1b2a1", elmts=[input.pop(0)])
        )

        t0, t1 = circuit.fp12_decompress_karabina_pt_I_c1b2_NZ(
            [xc0b1a0, xc0b1a1],
            [xc0b2a0, xc0b2a1],
            [xc1b0a0, xc1b0a1],
            [xc1b2a0, xc1b2a1],
        )
        circuit.extend_struct_output(structs.u384(name="t0a0", elmts=[t0[0]]))
        circuit.extend_struct_output(structs.u384(name="t0a1", elmts=[t0[1]]))
        circuit.extend_struct_output(structs.u384(name="t1a0", elmts=[t1[0]]))
        circuit.extend_struct_output(structs.u384(name="t1a1", elmts=[t1[1]]))

        return circuit


class E12TDecompressKarabinaPtIICircuit(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(
            "e12t_decomp_karabina_II", curve_id, auto_run, init_hash, compilation_mode
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(12)])

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = MillerTowerCircuit(self.name, self.curve_id, n_pairs=None)
        t0a0 = circuit.write_struct(structs.u384(name="t0a0", elmts=[input.pop(0)]))
        t0a1 = circuit.write_struct(structs.u384(name="t0a1", elmts=[input.pop(0)]))
        t1a0 = circuit.write_struct(structs.u384(name="t1a0", elmts=[input.pop(0)]))
        t1a1 = circuit.write_struct(structs.u384(name="t1a1", elmts=[input.pop(0)]))
        xc0b1a0 = circuit.write_struct(
            structs.u384(name="xc0b1a0", elmts=[input.pop(0)])
        )
        xc0b1a1 = circuit.write_struct(
            structs.u384(name="xc0b1a1", elmts=[input.pop(0)])
        )
        xc0b2a0 = circuit.write_struct(
            structs.u384(name="xc0b2a0", elmts=[input.pop(0)])
        )
        xc0b2a1 = circuit.write_struct(
            structs.u384(name="xc0b2a1", elmts=[input.pop(0)])
        )
        xc1b0a0 = circuit.write_struct(
            structs.u384(name="xc1b0a0", elmts=[input.pop(0)])
        )
        xc1b0a1 = circuit.write_struct(
            structs.u384(name="xc1b0a1", elmts=[input.pop(0)])
        )
        xc1b2a0 = circuit.write_struct(
            structs.u384(name="xc1b2a0", elmts=[input.pop(0)])
        )
        xc1b2a1 = circuit.write_struct(
            structs.u384(name="xc1b2a1", elmts=[input.pop(0)])
        )

        zc0b0, zc1b1 = circuit.fp12_decompress_karabina_pt_II(
            [t0a0, t0a1],
            [t1a0, t1a1],
            [xc0b1a0, xc0b1a1],
            [xc0b2a0, xc0b2a1],
            [xc1b0a0, xc1b0a1],
            [xc1b2a0, xc1b2a1],
        )
        circuit.extend_struct_output(structs.u384(name="zc0b0a0", elmts=[zc0b0[0]]))
        circuit.extend_struct_output(structs.u384(name="zc0b0a1", elmts=[zc0b0[1]]))
        circuit.extend_struct_output(structs.u384(name="zc1b1a0", elmts=[zc1b1[0]]))
        circuit.extend_struct_output(structs.u384(name="zc1b1a1", elmts=[zc1b1[1]]))

        return circuit



from garaga import garaga_rs
from garaga.algebra import ModuloCircuitElement, PyFelt
from garaga.definitions import BASE, N_LIMBS, STARK
from garaga.hints.io import bigint_split


def hades_permutation(s0: int, s1: int, s2: int) -> tuple[int, int, int]:
    r0, r1, r2 = garaga_rs.hades_permutation(
        (s0 % STARK).to_bytes(32, "big"),
        (s1 % STARK).to_bytes(32, "big"),
        (s2 % STARK).to_bytes(32, "big"),
    )
    return (
        int.from_bytes(r0, "big"),
        int.from_bytes(r1, "big"),
        int.from_bytes(r2, "big"),
    )


class CairoPoseidonTranscript:
    """
    The CairoPoseidonTranscript class mimics the behaviour of the Cairo functions hashing
    an array of u384 elements.
    """

    def __init__(self, init_hash: int) -> None:
        self.init_hash = init_hash
        self.s0, self.s1, self.s2 = hades_permutation(
            init_hash,
            0,
            1,
        )
        self.permutations_count = 1
        self.poseidon_ptr_indexes = []
        self.z = None

    @property
    def continuable_hash(self) -> int:
        return self.s0

    @property
    def RLC_coeff(self):
        """
        A function to retrieve the random linear combination coefficient after a permutation.
        Stores the index of the last permutation in the poseidon_ptr_indexes list, to be used to retrieve RLC coefficients later.
        """
        self.poseidon_ptr_indexes.append(self.permutations_count - 1)
        return self.s1

    def update_sponge_state(self, x, y):
        self.s0, self.s1, self.s2 = hades_permutation(self.s0 + x, self.s1 + y, self.s2)

    def hash_element(self, x: PyFelt | ModuloCircuitElement, debug: bool = False):
        # print(f"Will Hash PYTHON {hex(x.value)}")
        limbs = bigint_split(x, N_LIMBS, BASE)
        if debug:
            print(f"limbs : {limbs}")
        self.s0, self.s1, self.s2 = hades_permutation(
            self.s0 + limbs[0] + (BASE) * limbs[1],
            self.s1 + limbs[2] + (BASE) * limbs[3],
            self.s2,
        )
        self.permutations_count += 1

        return self.s0, self.s1

    def hash_u256(self, x: PyFelt | int):
        assert isinstance(x, (PyFelt, int))
        if isinstance(x, PyFelt):
            x = x.value
        assert 0 <= x < 2**256
        low, high = bigint_split(x, 2, 2**128)
        self.s0, self.s1, self.s2 = hades_permutation(
            self.s0 + low, self.s1 + high, self.s2
        )
        self.permutations_count += 1
        return self.s0

    def hash_u128(self, x: PyFelt | int):
        assert isinstance(x, (PyFelt, int))
        if isinstance(x, PyFelt):
            x = x.value
        assert 0 <= x < 2**128
        self.s0, self.s1, self.s2 = hades_permutation(self.s0 + x, self.s1, self.s2)
        self.permutations_count += 1
        return self.s0

    def hash_u256_multi(self, X: list[PyFelt | int]):
        for x in X:
            self.hash_u256(x)
        return self.s0

    def hash_limbs_multi(
        self,
        X: list[PyFelt | ModuloCircuitElement],
        sparsity: list[int] | None = None,
        debug: bool = False,
    ):
        if sparsity:
            X = [x for i, x in enumerate(X) if sparsity[i] != 0]
        for X_elem in X:
            if debug:
                print(f"\t s0 : {self.s0}")
            self.hash_element(X_elem, debug=debug)
        return None


if __name__ == "__main__":
    import random

    random.seed(0)

    def gen_cairo_test_vectors(n_elmts: int):
        elmts = [PyFelt(random.randint(0, 2**384 - 1), 2**384) for _ in range(n_elmts)]
        transcript = CairoPoseidonTranscript(init_hash=0)
        for elmt in elmts:
            transcript.hash_element(elmt)
        expected_res = transcript.continuable_hash

        code = f"""
    #[test]
    fn test_hash_u384_{n_elmts}() {{
    """
        code += "// Auto-generated from garaga/poseidon_transcript.py\n"
        code += "let transcript: Array<u384> = array!["
        for elmt in elmts:
            limbs = bigint_split(elmt.value, N_LIMBS, BASE)
            code += f"u384 {{ limb0: {limbs[0]}, limb1: {limbs[1]}, limb2: {limbs[2]}, limb3: {limbs[3]} }},"
        code += "];\n"
        code += f"let expected_res: felt252 = {expected_res};\n"
        code += "let res = hash_u384_transcript(transcript.span(), 0);\n"
        code += "assert_eq!(res, expected_res);\n"
        code += "}\n\n"
        return code

    print(gen_cairo_test_vectors(1))
    print(gen_cairo_test_vectors(2))
    print(gen_cairo_test_vectors(3))
